#!/usr/bin/env python3
# Copyright (c) 2020-2024 Valve Corporation
# Copyright (c) 2020-2024 LunarG, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Script to determine if source code in Pull Request is properly formatted.
#
# This script checks for:
#   -- clang-format errors in the PR source code
#   -- out-of-date copyrights in PR source files
#   -- improperly formatted commit messages (using the function above)
#   -- assigning stype instead of using vku::InitStruct
#
# Notes:
#    Exits with non 0 exit code if formatting is needed.
#    Requires python3 to run correctly
#    In standalone mode (outside of CI), changes must be rebased on main
#        to get meaningful and complete results

import os
import argparse
import re
import subprocess
from subprocess import check_output
from argparse import RawDescriptionHelpFormatter

#
#
# Color print routine, takes a string matching a txtcolor above and the output string, resets color upon exit
def CPrint(msg_type, msg_string):
    txtcolors = {'HELP_MSG':    '\033[0;36m',
                 'SUCCESS_MSG': '\033[1;32m',
                 'CONTENT':     '\033[1;39m',
                 'ERR_MSG':     '\033[1;31m',
                 'NO_COLOR':    '\033[0m'}
    print(txtcolors.get(msg_type, txtcolors['NO_COLOR']) + msg_string + txtcolors['NO_COLOR'])
#
#
# Check clang-formatting of source code diff
def VerifyClangFormatSource(commit, target_files):
    target_refspec = f'{commit}^...{commit}'
    retval = 0
    good_file_pattern = re.compile('.*\\.(cpp|cc|c\+\+|cxx|c|h|hpp)$')
    diff_files_list = [item for item in target_files if good_file_pattern.search(item)]
    diff_files = ' '.join([str(elem) for elem in diff_files_list])
    retval = 0
    if diff_files != '':
        git_diff = subprocess.Popen(('git', 'diff', '-U0', target_refspec, '--', diff_files), stdout=subprocess.PIPE)
        diff_files_data = subprocess.check_output(('python3', './scripts/clang-format-diff.py', '-p1', '-style=file'), stdin=git_diff.stdout)
        diff_files_data = diff_files_data.decode('utf-8')
        if diff_files_data != '':
            CPrint('ERR_MSG', "\nFound formatting errors!")
            CPrint('CONTENT', "\n" + diff_files_data)
            retval = 1
    return retval
#
#
# Check copyright dates for modified files
def VerifyCopyrights(commit, target_files):
    retval = 0
    is_lunarg_author = False
    authors = check_output(['git', 'log', '-n' , '1', '--format=%ae', commit])
    for author in authors.split(b'\n'):
        if author.endswith(b'@lunarg.com'):
            is_lunarg_author = True
            break

    if not is_lunarg_author:
        return 0

    # Handle year changes by respecting when the author wrote the code, rather
    # the day the script runs. This isn't exactly right yet, because really
    # we should evaluate it commit's files against that commit's date.
    commit_year = None
    # get all the author dates in YYYY-MM-DD format
    commit_dates = check_output(['git', 'log', '-n', '1', '--format=%as', commit])
    for cd in commit_dates.split(b'\n'):
        if len(cd) == 0:
            continue
        year = cd.split(b'-')[0]
        if not commit_year or int(commit_year) < int(year):
            commit_year = year.decode('utf-8')
    for file in target_files:
        if file is None or not os.path.isfile(file):
            continue
        for company in ["LunarG", "Valve"]:
            # Capture the last year on the line as a separate match. It should be the highest (or only year of the range)
            copyright_match = re.search('Copyright .*(\d{4}) ' + company, open(file, encoding="utf-8", errors='ignore').read(1024))
            if copyright_match:
                copyright_year = copyright_match.group(1)
                if int(commit_year) > int(copyright_year):
                    msg = 'Change written in {} but copyright ends in {}.'.format(commit_year, copyright_year)
                    CPrint('ERR_MSG', '\n' + file + ' has an out-of-date ' + company + ' copyright notice. ' + msg)
                    retval = 1
    return retval
#
#
# Check commit message formats for commits in this PR/Branch
def VerifyCommitMessageFormat(commit, target_files):
    retval = 0

    # Construct correct commit list
    commit_text= check_output(['git', 'log', '-n', '1', '--pretty=format:%B', commit]).decode('utf-8')
    if commit_text is None:
        return retval

    msg_cur_line = 0
    msg_prev_line = ''
    for msg_line_text in commit_text.splitlines():
        msg_cur_line += 1
        line_length = len(msg_line_text)

        if msg_cur_line == 1:
            # Enforce subject line must be 64 chars or less
            if line_length > 64:
                CPrint('ERR_MSG', "The following subject line exceeds 64 characters in length.")
                CPrint('CONTENT', "     '" + msg_line_text + "'\n")
                retval = 1
            # Output error if last char of subject line is not alpha-numeric
            if msg_line_text[-1] in '.,':
                CPrint('ERR_MSG', "For the following commit, the last character of the subject line must not be a period or comma.")
                CPrint('CONTENT', "     '" + msg_line_text + "'\n")
                retval = 1
            # Output error if subject line doesn't start with 'module: '
            if 'Revert' not in msg_line_text:
                module_name = msg_line_text.split(' ')[0]
                if module_name[-1] != ':':
                    CPrint('ERR_MSG', "The following subject line must start with a single word specifying the functional area of the change, followed by a colon and space.")
                    CPrint('ERR_MSG', "e.g., 'layers: Subject line here' or 'corechecks: Fix off-by-one error in ValidateFences'.")
                    CPrint('ERR_MSG', "Other common module names include layers, build, cmake, tests, docs, scripts, stateless, gpu, syncval, practices, etc.")
                    CPrint('CONTENT', "     '" + msg_line_text + "'\n")
                    retval = 1
                else:
                    # Check if first character after the colon is lower-case
                    subject_body = msg_line_text.split(': ')[1]
                    if not subject_body[0].isupper():
                        CPrint('ERR_MSG', "The first word of the subject line after the ':' character must be capitalized.")
                        CPrint('CONTENT', "     '" + msg_line_text + "'\n")
                        retval = 1
            # Check that first character of subject line is not capitalized
            if msg_line_text[0].isupper():
                CPrint('ERR_MSG', "The first word of the subject line must be lower case.")
                CPrint('CONTENT', "     '" + msg_line_text + "'\n")
                retval = 1
        elif msg_cur_line == 2:
            # Commit message must have a blank line between subject and body
            if line_length != 0:
                CPrint('ERR_MSG', "The following subject line must be followed by a blank line.")
                CPrint('CONTENT', "     '" + msg_prev_line + "'\n")
                retval = 1
        else:
            # Lines in a commit message body must be less than 72 characters in length (but give some slack)
            if line_length > 76:
                CPrint('ERR_MSG', "The following commit message body line exceeds the 72 character limit.")
                CPrint('CONTENT', "     '" + msg_line_text + "'\n")
                retval = 1
        msg_prev_line = msg_line_text
    if retval != 0:
        CPrint('HELP_MSG', "Commit Message Format Requirements:")
        CPrint('HELP_MSG', "-----------------------------------")
        CPrint('HELP_MSG', "o  Subject lines must be <= 64 characters in length")
        CPrint('HELP_MSG', "o  Subject lines must start with a module keyword which is lower-case and followed by a colon and a space")
        CPrint('HELP_MSG', "o  The first word following the colon must be capitalized and the subject line must not end in a '.'")
        CPrint('HELP_MSG', "o  The subject line must be followed by a blank line")
        CPrint('HELP_MSG', "o  The commit description must be <= 72 characters in width\n")
        CPrint('HELP_MSG', "Examples:")
        CPrint('HELP_MSG', "---------")
        CPrint('HELP_MSG', "     build: Fix Vulkan header/registry detection for SDK")
        CPrint('HELP_MSG', "     tests: Fix QueryPerformanceIncompletePasses stride usage")
        CPrint('HELP_MSG', "Refer to CONTRIBUTING.md for additional detail")
    return retval

#
#
# Check for test code assigning sType instead of using vku::InitStruc in this PR/Branch
def VerifyTypeAssign(commit, target_files):
    retval = 0
    target_refspec = f'{commit}^...{commit}'

    test_files_list = [item for item in target_files if item.startswith('tests/')]
    test_files = ' '.join([str(elem) for elem in test_files_list])
    if not test_files:
        return 0
    test_diff = subprocess.Popen(('git', 'diff', '-U0', target_refspec, '--', test_files), stdout=subprocess.PIPE)
    stdout, stderr = test_diff.communicate()
    stdout = stdout.decode('utf-8')
    stype_regex = re.compile(r'\.sType\s*=')
    on_regex = re.compile(r'stype-check\s*on')
    off_regex = re.compile(r'stype-check\s*off')
    checking = True
    for line in stdout.split('\n'):
        if not line.startswith('-'):
            if checking:
                if off_regex.search(line, re.IGNORECASE):
                    checking = False
                elif stype_regex.search(line):
                    CPrint('ERR_MSG', "Test assigning sType instead of using vku::InitStruct")
                    CPrint('ERR_MSG', "If this is a case where vku::InitStruct cannot be used, //stype-check off can be used to turn off sType checking")
                    CPrint('CONTENT', "     '" + line + "'\n")
                    retval = 1
            else:
                if on_regex.search(line, re.IGNORECASE):
                    checking = True
    return retval
#
#
# Entrypoint
def main():
    DEFAULT_REFSPEC = 'origin/main'

    parser = argparse.ArgumentParser(description='''Usage: python ./scripts/check_code_format.py
    - Reqires python3 and clang-format
    - Run script in repo root
    - May produce inaccurate clang-format results if local branch is not rebased on the TARGET_REFSPEC
    ''', formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('--target-refspec', metavar='TARGET_REFSPEC', type=str, dest='target_refspec', help = 'Refspec to '
        + 'diff against (default is origin/main)', default=DEFAULT_REFSPEC)
    parser.add_argument('--base-refspec', metavar='BASE_REFSPEC', type=str, dest='base_refspec', help = 'Base refspec to '
        + ' compare (default is HEAD)', default='HEAD')
    parser.add_argument('--fetch-main', dest='fetch_main', action='store_true', help='Fetch the main branch first.'
        + ' Useful with --target-refspec=FETCH_HEAD to compare against what is currently on main')
    args = parser.parse_args()

    if os.path.isfile('check_code_format.py'):
        os.chdir('..')

    target_refspec = args.target_refspec
    base_refspec = args.base_refspec

    if args.fetch_main:
        print('Fetching main branch...')
        subprocess.check_call(['git', 'fetch', 'origin', 'main'])

    # Check if this is a merge commit
    commit_parents = check_output(['git', 'rev-list', '--parents', '-n', '1', 'HEAD'])
    if len(commit_parents.split(b' ')) > 2:
        # If this is a merge commit, this is a PR being built, and has been merged into main for testing.
        # The first parent (HEAD^) is going to be main, the second parent (HEAD^2) is going to be the PR commit.
        # TODO (ncesario) We should *ONLY* get here when on github CI, building a PR. Should probably print a
        #      warning if this happens locally.
        target_refspec = 'HEAD^'
        base_refspec = 'HEAD^2'

    orig_branch = check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).decode('utf-8').splitlines()[0]
    if orig_branch == 'HEAD':
        orig_branch = check_output(['git', 'rev-parse', 'HEAD']).decode('utf-8').splitlines()[0]

    commits = check_output(['git', 'log', '--format=%h', f'{base_refspec}...{target_refspec}']).split(b'\n')
    commits.reverse()

    # Run code format check on each commit in a PR so that we ensure that each commit is correct.
    failure = 0
    for c in commits:
        if len(c) == 0:
            continue

        commit = c.decode('utf-8')
        diff_range = f'{commit}^...{commit}'

        commit_message = check_output(['git', 'log', '--pretty="%h %s"', diff_range])
        CPrint('CONTENT', "\nChecking commit: " + commit_message.decode('utf-8'))

        subprocess.run(['git', 'checkout', '-q', commit])

        # Get list of files involved in this commit
        target_files_data = subprocess.check_output(['git', 'log', '-n', '1', '--name-only', commit])
        target_files = target_files_data.decode('utf-8')
        target_files = target_files.split("\n")

        # Skip checking dependabot commits
        authors = subprocess.check_output(['git', 'log', '-n' , '1', '--format=%ae', commit]).decode('utf-8')
        if "dependabot" in authors:
            continue

        failure |= VerifyClangFormatSource(commit, target_files)
        failure |= VerifyCopyrights(commit, target_files)
        failure |= VerifyCommitMessageFormat(commit, target_files)
        failure |= VerifyTypeAssign(commit, target_files)

    subprocess.run(['git', 'checkout', '-q', orig_branch])

    if failure:
        CPrint('ERR_MSG', "One or more format checks failed.\n")
        exit(1)

    CPrint('SUCCESS_MSG', "All format checks passed.\n")

if __name__ == '__main__':
  main()



#!/usr/bin/env python3
# Copyright (c) 2020-2024 Valve Corporation
# Copyright (c) 2020-2024 LunarG, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import sys
import os
import argparse
import common_ci

# Where all artifacts will ultimately be placed under
CI_BUILD_DIR = common_ci.RepoRelative('build-ci')
# Where all dependencies will be installed under
CI_EXTERNAL_DIR = f'{CI_BUILD_DIR}/external'

# Prepare the layer for testing
def BuildCDL(config, cmake_args, build_tests, sanitize):
    print("Log CMake version")
    cmake_ver_cmd = 'cmake --version'
    common_ci.RunShellCmd(cmake_ver_cmd)

    SRC_DIR = common_ci.PROJECT_SRC_DIR
    BUILD_DIR = f'{CI_BUILD_DIR}/cdl'



    print("Configure CDL")
    cmake_cmd = f'cmake -S {SRC_DIR} -B {BUILD_DIR}'
    cmake_cmd += f' -D CMAKE_BUILD_TYPE={config}'
    cmake_cmd += f' -D BUILD_TESTS={build_tests}'
    cmake_cmd += f' -D UPDATE_DEPS=ON -D UPDATE_DEPS_DIR={CI_EXTERNAL_DIR}'

    if cmake_args:
         cmake_cmd += f' {cmake_args}'


    if sanitize == 'none':
        cmake_cmd += ' -D BUILD_WERROR=ON'
    else:
        os.environ['CFLAGS'] = f'-fsanitize={sanitize}'
        os.environ['CXXFLAGS'] = f'-fsanitize={sanitize}'
        os.environ['LDFLAGS'] = f'-fsanitize={sanitize}'
        cmake_cmd += ' -D BUILD_WERROR=OFF'
    print('env', os.environ)
    print('cmake_cmd', cmake_cmd)
    common_ci.RunShellCmd(cmake_cmd)

    print("Build CDL")
    build_cmd = f'cmake --build {BUILD_DIR}'
    common_ci.RunShellCmd(build_cmd)


#
# Module Entrypoint
def Build(args):
    config = args.configuration
    sanitize = args.sanitize

    # Since this script uses Ninja to build Windows users need to be in a developer command prompt.
    if common_ci.IsWindows():
        # This environment variable is arbitrary. I just picked one set by the developer command prompt.
        if "VSCMD_ARG_TGT_ARCH" not in os.environ:
            print("This script must be invoked in a developer command prompt!")
            sys.exit(1)

    try:
        BuildCDL(config = config, cmake_args = args.cmake, build_tests = "ON", sanitize = sanitize)

    except subprocess.CalledProcessError as proc_error:
        print('Command "%s" failed with return code %s' % (' '.join(proc_error.cmd), proc_error.returncode))
        sys.exit(proc_error.returncode)
    except Exception as unknown_error:
        print('An unknown error occured: %s', unknown_error)
        sys.exit(1)

    sys.exit(0)

#
# Run the Layer Validation Tests
def RunTests(args):
    print("Run Tests using Mock ICD")

    lvt_env = dict(os.environ)
    lvt_cmd = os.path.join(CI_BUILD_DIR, 'cdl', 'tests', 'cdl_tests')

    common_ci.RunShellCmd(lvt_cmd, env=lvt_env)

def Test(args):
    try:
        RunTests(args)

    except subprocess.CalledProcessError as proc_error:
        print('Command "%s" failed with return code %s' % (' '.join(proc_error.cmd), proc_error.returncode))
        sys.exit(proc_error.returncode)
    except Exception as unknown_error:
        print('An unknown error occured: %s', unknown_error)
        sys.exit(1)

    sys.exit(0)

if __name__ == '__main__':
    configs = ('release', 'debug')
    default_config = configs[0]

    sanitize_opts = ('none', 'address', 'thread')

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-c', '--config', dest='configuration',
        metavar='CONFIG', action='store',
        choices=configs, default=default_config,
        help='Build target configuration. Can be one of: {0}'.format(
            ', '.join(configs)))
    parser.add_argument(
        '-s', '--sanitize', dest='sanitize',
        metavar='SANITIZE', action='store',
        choices=sanitize_opts, default='none',
        help='-fsanitize setting. Can be one of: {0}'.format(', '.join(sanitize_opts)))
    parser.add_argument(
        '--cmake', dest='cmake',
        metavar='CMAKE', type=str,
        default='', help='Additional args to pass to cmake')
    parser.add_argument(
        '--build', dest='build',
        action='store_true', help='Build the layers')
    parser.add_argument(
        '--test', dest='test',
        action='store_true', help='Tests the layers')

    args = parser.parse_args()

    if (args.build):
        Build(args)
    if (args.test):
        Test(args)



#!/usr/bin/python3 -i
#
# Copyright (c) 2023 LunarG, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from generators.cdl_base_generator import CdlBaseOutputGenerator
#
# CommandCommonOutputGenerator - Generate command output and storage
class CommandCommonOutputGenerator(CdlBaseOutputGenerator):
    def __init__(self):
        CdlBaseOutputGenerator.__init__(self)

    #
    # Called at beginning of processing as file is opened
    def generate(self):
        file_start = self.GenerateFileStart(os.path.basename(__file__))
        self.write(file_start)

        if self.filename == 'command_common.h':
            self.generateHeader()
        elif self.filename == 'command_common.cpp':
            self.generateSource()
        else:
            self.write(f'\nFile name {self.filename} has no code to generate\n')

        self.write(self.GenerateFileEnd())

    def generateHeader(self):
        self.write('\n#pragma once\n')
        self.write('#include <cstring>')
        self.write('#include <iostream>')
        self.write('#include <string>')
        self.write('#include <vector>')
        self.write('#include <vulkan/vulkan.h>')

        #
        # List the enum for the commands
        out = []
        out.append('''
// Enumerate commands that have one parameter of type VkCommandBuffer.
struct Command
{
  enum Type {
    kUnknown,\n\n''')
        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append(f'    k{vkcommand.name[2:]},\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
        out.append('''}; // enum Type

  public:
  static const char *GetCommandName(const Command &cmd);

  Type type;
  uint32_t id;
  void *parameters;
  std::vector<std::string> labels;
}; // struct Command

// Define structs for command parameters
''')

        self.write("".join(out))

        #
        # For each CMD_TYPE give a string name
        out = []
        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append(f'struct {vkcommand.name[2:]}Args {{\n')
            for vkparam in vkcommand.params:
                param_stripped = vkparam.cDeclaration.strip()
                if vkparam.const and not vkparam.pointer:
                    param_decl = param_stripped[6:]
                else:
                    param_decl = param_stripped
                out.append(f'    {param_decl};\n')
            out.append('};\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')

        self.write("".join(out))

    def generateSource(self):
        self.write('#include "command_common.h"')

        out = []
        out.append('''
// Returns the name of given command. This only cares for commands that have one
// parameter of type VkCommandBuffer.
const char *Command::GetCommandName(const Command &cmd) {
  switch (cmd.type) {
  default:
  case Command::Type::kUnknown:
    return "Unknown";

''')
        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append(f'  case Command::Type::k{vkcommand.name[2:]}:\n')
            out.append(f'    return "{vkcommand.name}";\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])

        out.append('  } // switch (cmd.type) \n')
        out.append('} // Command::GetCommandName()\n')
        self.write("".join(out))



#!/usr/bin/python3 -i
#
# Copyright (c) 2015-2024 The Khronos Group Inc.
# Copyright (c) 2015-2024 Valve Corporation
# Copyright (c) 2015-2024 LunarG, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import subprocess
import platform

# Use Ninja for all platforms for performance/simplicity
os.environ['CMAKE_GENERATOR'] = "Ninja"

# helper to define paths relative to the repo root
def RepoRelative(path):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), '..', path))

# Points to the directory containing the top level CMakeLists.txt
PROJECT_SRC_DIR = os.path.abspath(os.path.join(os.path.split(os.path.abspath(__file__))[0], '..'))
if not os.path.isfile(f'{PROJECT_SRC_DIR}/CMakeLists.txt'):
    print(f'PROJECT_SRC_DIR invalid! {PROJECT_SRC_DIR}')
    sys.exit(1)

# Returns true if we are running in GitHub actions
# https://docs.github.com/en/actions/learn-github-actions/variables#default-environment-variables
def IsGHA():
    if 'GITHUB_ACTION' in os.environ:
        return True
    return False

# Runs a command in a directory and returns its return code.
# Directory is project root by default, or a relative path from project root
def RunShellCmd(command, start_dir = PROJECT_SRC_DIR, env=None, verbose=False):
    # Flush stdout here. Helps when debugging on CI.
    sys.stdout.flush()

    if start_dir != PROJECT_SRC_DIR:
        start_dir = RepoRelative(start_dir)
    cmd_list = command.split(" ")

    # Helps a lot when debugging CI issues
    if IsGHA():
        verbose = True

    if verbose:
        print(f'CICMD({cmd_list}, env={env})')
    subprocess.check_call(cmd_list, cwd=start_dir, env=env)

#
# Check if the system is Windows
def IsWindows(): return 'windows' == platform.system().lower()



#!/usr/bin/python3 -i
#
# Copyright (c) 2023-2025 LunarG, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from base_generator import BaseGenerator
from enum import IntEnum

# class syntax
class InterceptFlag(IntEnum):
    PRE = (1 << 0)
    POST = (1 << 1)
    NORMAL = (PRE | POST)
    # Single hook generated rather than Pre/Post. This hook must dispatch the call
    OVERRIDE = (1 << 2)
    # Intercept* function is hand coded
    CUSTOM = (1 << 3)

#NOTE: anything named vkCmd* is implicitly a NORMAL intercept so it doesn't need to be
# in the table below
intercept_functions = {
        'vkAcquireNextImageKHR' : InterceptFlag.NORMAL,
        'vkAllocateCommandBuffers': InterceptFlag.NORMAL,
        'vkCreateCommandPool': InterceptFlag.NORMAL,
        'vkCreateComputePipelines': InterceptFlag.POST,
        'vkCreateDebugReportCallbackEXT': InterceptFlag.NORMAL,
        'vkCreateDebugUtilsMessengerEXT': InterceptFlag.NORMAL,
        'vkCreateDevice': InterceptFlag.CUSTOM | InterceptFlag.POST,
        'vkCreateGraphicsPipelines': InterceptFlag.POST,
        'vkCreateInstance': InterceptFlag.CUSTOM | InterceptFlag.POST,
        'vkCreateSemaphore': InterceptFlag.NORMAL,
        'vkCreateShaderModule': InterceptFlag.POST,
        'vkDebugMarkerSetObjectNameEXT': InterceptFlag.NORMAL,
        'vkDestroyCommandPool': InterceptFlag.NORMAL,
        'vkDestroyDebugReportCallbackEXT': InterceptFlag.NORMAL,
        'vkDestroyDebugUtilsMessengerEXT': InterceptFlag.NORMAL,
        'vkDestroyDevice': InterceptFlag.CUSTOM | InterceptFlag.PRE,
        'vkDestroyInstance': InterceptFlag.CUSTOM | InterceptFlag.PRE,
        'vkDestroyPipeline': InterceptFlag.NORMAL,
        'vkDestroySemaphore': InterceptFlag.NORMAL,
        'vkDestroyShaderModule': InterceptFlag.PRE,
        'vkDeviceWaitIdle': InterceptFlag.NORMAL,
        'vkEnumerateDeviceExtensionProperties': InterceptFlag.CUSTOM,
        'vkEnumerateDeviceLayerProperties': InterceptFlag.CUSTOM,
        'vkEnumerateInstanceExtensionProperties': InterceptFlag.CUSTOM,
        'vkEnumerateInstanceLayerProperties': InterceptFlag.CUSTOM,
        'vkFreeCommandBuffers': InterceptFlag.NORMAL,
        'vkGetDeviceProcAddr': InterceptFlag.CUSTOM,
        'vkGetDeviceQueue': InterceptFlag.POST,
        'vkGetDeviceQueue2': InterceptFlag.POST,
        'vkGetFenceStatus': InterceptFlag.NORMAL,
        'vkGetInstanceProcAddr': InterceptFlag.CUSTOM,
        'vkGetPhysicalDeviceToolProperties': InterceptFlag.CUSTOM,
        'vkGetPhysicalDeviceToolPropertiesEXT': InterceptFlag.CUSTOM,
        'vkGetQueryPoolResults': InterceptFlag.NORMAL,
        'vkGetSemaphoreCounterValue': InterceptFlag.NORMAL,
        'vkGetSemaphoreCounterValueKHR': InterceptFlag.NORMAL,
        'vkQueueBindSparse': InterceptFlag.OVERRIDE,
        'vkQueuePresentKHR': InterceptFlag.NORMAL,
        'vkQueueSubmit': InterceptFlag.OVERRIDE,
        'vkQueueSubmit2': InterceptFlag.OVERRIDE,
        'vkQueueSubmit2KHR': InterceptFlag.OVERRIDE,
        'vkQueueWaitIdle': InterceptFlag.NORMAL,
        'vkResetCommandPool': InterceptFlag.NORMAL,
        'vkSetDebugUtilsObjectNameEXT': InterceptFlag.NORMAL,
        'vkSignalSemaphore': InterceptFlag.NORMAL,
        'vkSignalSemaphoreKHR': InterceptFlag.NORMAL,
        'vkWaitForFences': InterceptFlag.NORMAL,
        'vkWaitSemaphores': InterceptFlag.NORMAL,
        'vkWaitSemaphoresKHR': InterceptFlag.NORMAL,
}

namespace = 'crash_diagnostic_layer'

#
# CdlBaseOutputGenerator - Base Generator for the whole Crash Diagnostic Layer
class CdlBaseOutputGenerator(BaseGenerator):
    def __init__(self):
        BaseGenerator.__init__(self)
        self.intercept_functions = intercept_functions

    def GenerateFileStart(self, filename):
        file_start = f'''
/***************************************************************************
*
* Copyright (C) 2021 Google Inc.
* Copyright (c) 2023-2025 LunarG, Inc.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
* *** ---- WARNING! ----
* ***   THIS FILE IS GENERATED - DO NOT EDIT
* ***   Update source file {filename} for modifications
* *** ---- WARNING! ----
*
****************************************************************************/\n

// NOLINTBEGIN
'''
        return file_start

    def GenerateFileEnd(self):
        file_end = '\n// NOLINTEND\n'
        return file_end

    def GenerateNamespaceBegin(self):
        namespace_text = f'\nnamespace {namespace} {{\n'
        return namespace_text

    def GenerateNamespaceEnd(self):
        namespace_text = f'\n}} // namespace {namespace}\n'
        return namespace_text

    def CreateNamespaceReference(self, item):
        namespace_text = f'{namespace}::{item}'
        return namespace_text

    def CommandBufferCall(self, command):
        intercept = (command.name.startswith('vkCmd') or
                     command.params[0].type == 'VkCommandBuffer')
        return intercept

    def NeedsIntercept(self, command):
        intercept = (self.CommandBufferCall(command) or
                        intercept_functions.get(command.name, None) is not None)
        return intercept

    def InterceptPreCommand(self, command):
        attrs = intercept_functions.get(command.name, None)
        if attrs is not None:
            return (attrs & InterceptFlag.PRE) != 0
        return self.CommandBufferCall(command)

    def InterceptPostCommand(self, command):
        attrs = intercept_functions.get(command.name, None)
        if attrs is not None:
            return (attrs & InterceptFlag.POST) != 0
        return self.CommandBufferCall(command)

    def InterceptGenerateSource(self, command):
        attrs = intercept_functions.get(command.name, None)
        if attrs is not None:
            return (attrs & InterceptFlag.CUSTOM) == 0
        return self.CommandBufferCall(command)

    def InterceptOverrideCommand(self, command):
        attrs = intercept_functions.get(command.name, None)
        return attrs is not None and (attrs & InterceptFlag.OVERRIDE) != 0

    def InstanceCommand(self, command):
        return command.instance or command.params[0].type == 'VkPhysicalDevice'

    def CommandHasReturn(self, command):
        return command.returnType is not None and command.returnType != 'void'

    def GetDefaultReturnValue(self, return_type):
        return_string = ''
        if return_type == 'VkResult':
            return_string = 'VK_SUCCESS'
        elif return_type in self.vk.structs.values():
            return_string = 'nullptr'
        elif 'int' in return_type and '64' not in return_type:
            return_string = '0'
        elif 'int64' in return_type or 'Address' in return_type or 'Size' in return_type:
            return_string = '0ULL'
        elif return_type == 'VkBool32':
            return_string = 'VK_FALSE'
        elif 'PFN_' in return_type:
            return_string = 'nullptr'
        elif 'PFN_' in return_type:
            return_string = 'UNKNOWN_VALUE'
        return return_string



#!/usr/bin/python3 -i
#
# Copyright (c) 2023-2024 LunarG, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from generators.cdl_base_generator import CdlBaseOutputGenerator

custom_functions = (
    'vkCmdBeginRendering',
    'vkCmdEndRendering',
)

default_instrumented_functions = (
    'vkCmdDispatch',
    'vkCmdDispatchIndirect',
    'vkCmdDraw',
    'vkCmdDrawIndexed',
    'vkCmdDrawIndirect',
    'vkCmdDrawIndexedIndirect',
    'vkCmdCopyBuffer',
    'vkCmdCopyBufferToImage',
    'vkCmdCopyImage',
    'vkCmdCopyImageToBuffer',
    'vkCmdBindPipeline',
    'vkCmdExecuteCommands',
    'vkCmdPipelineBarrier',
    'vkCmdSetEvent',
    'vkCmdResetEvent',
    'vkCmdWaitEvents',
    'vkCmdDebugMarkerBeginEXT',
    'vkCmdDebugMarkerEndEXT',
    'vkCmdDebugMarkerInsertEXT',
    'vkCmdBeginDebugUtilsLabelEXT',
    'vkCmdEndDebugUtilsLabelEXT',
    'vkCmdInsertDebugUtilsLabelEXT',
    'vkCmdTraceRaysKHR',
    'vkCmdTraceRaysIndirectKHR',
    'vkCmdTraceRaysNV',
    'vkCmdBuildAccelerationStructuresKHR',
    'vkCmdBuildAccelerationStructuresIndirectKHR'
    'vkCmdBuildAccelerationStructuresNV',
)

#
# InterceptCommandsOutputGenerator - Generate the dispatch tables
class CommandPrePostGenerator(CdlBaseOutputGenerator):
    def __init__(self):
        CdlBaseOutputGenerator.__init__(self)

    #
    # Called at beginning of processing as file is opened
    def generate(self):
        file_start = self.GenerateFileStart(os.path.basename(__file__))
        self.write(file_start)

        if self.filename == 'command.h.inc':
            self.generateCommandsHeader()
        elif self.filename == 'command.cpp.inc':
            self.generateCommandsSource()
        else:
            self.write(f'\nFile name {self.filename} has no code to generate\n')

        self.write(self.GenerateFileEnd())

    def generateCommandsHeader(self):
        out = []
        for vkcommand in filter(lambda x: x.name.startswith('vkCmd') and x.name not in custom_functions, self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            post_func_decl = vkcommand.cPrototype.replace('VKAPI_ATTR ', '').replace('VKAPI_CALL ', '').replace(' vk', ' Post', 1)
            pre_func_decl = post_func_decl.replace('Post', 'Pre', 1)
            if self.CommandHasReturn(vkcommand):
                post_func_decl = post_func_decl.replace(')', f', {vkcommand.returnType} result)', 1)
            out.append(f'{pre_func_decl}\n')
            out.append(f'{post_func_decl}\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')
        self.write("".join(out))

    def generateCommandsSource(self):
        out = []
        for vkcommand in filter(lambda x: x.name.startswith('vkCmd') and x.name not in custom_functions, self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            post_func_decl = vkcommand.cPrototype.replace('VKAPI_ATTR ', '').replace('VKAPI_CALL ', '').replace(';', ' {').replace(' vk', ' CommandBuffer::Post', 1)
            pre_func_decl = post_func_decl.replace('Post', 'Pre', 1)
            func_call = f'  tracker_.{vkcommand.name[2:]}('
            count = 0
            for vkparam in vkcommand.params:
                if count != 0:
                    func_call += ', '
                func_call += vkparam.name
                count += 1
            func_call += ');'

            tracker_call = '  WriteCommandBeginCheckpoint(tracker_.GetCommands().back().id);'

            out.append(f'{pre_func_decl}\n')
            out.append(f'{func_call}\n')
            if vkcommand.name not in default_instrumented_functions:
                out.append('  if (instrument_all_commands_)\n')
                out.append('  ')
            out.append(f'{tracker_call}\n')

            if self.CommandHasReturn(vkcommand):
                default_value = self.GetDefaultReturnValue(vkcommand.returnType)
                out.append(f'  return {default_value};\n')

            out.append('}\n')

            if self.CommandHasReturn(vkcommand):
                post_func_decl = post_func_decl.replace(')', f',\n    {vkcommand.returnType}                                    result)')
            tracker_call = '  WriteCommandEndCheckpoint(tracker_.GetCommands().back().id);'

            out.append(f'{post_func_decl}\n')
            if vkcommand.name not in default_instrumented_functions:
                out.append('  if (instrument_all_commands_)\n')
                out.append('  ')
            out.append(f'{tracker_call}\n')
            if self.CommandHasReturn(vkcommand):
                out.append('  return result;\n')
            out.append('}\n')

            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')
        self.write("".join(out))




#!/usr/bin/python3 -i
#
# Copyright (c) 2023-2025 LunarG, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from vulkan_object import (Queues, CommandScope)
from generators.cdl_base_generator import CdlBaseOutputGenerator
#
# DispatchOutputGenerator - Generate the dispatch tables
class DispatchOutputGenerator(CdlBaseOutputGenerator):
    def __init__(self):
        CdlBaseOutputGenerator.__init__(self)

    #
    # Called at beginning of processing as file is opened
    def generate(self):
        file_start = self.GenerateFileStart(os.path.basename(__file__))
        self.write(file_start)

        if self.filename == 'dispatch.h':
            self.generateHeader()
        elif self.filename == 'dispatch.cpp':
            self.generateSource()
        else:
            self.write(f'\nFile name {self.filename} has no code to generate\n')

        self.write(self.GenerateFileEnd())

    def generateHeader(self):
        self.write('\n#pragma once\n')
        self.write('#include <vulkan/vk_layer.h>')
        self.write('#include <vulkan/vulkan.h>')
        self.write(self.GenerateNamespaceBegin())

        out = []
        out.append('\nstruct InstanceDispatchTable {\n')
        for vkcommand in filter(lambda x: self.InstanceCommand(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append(f'  PFN_{vkcommand.name} {vkcommand.name[2:]};\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
        out.append('};\n\n')

        out.append('struct DeviceDispatchTable {\n')
        for vkcommand in filter(lambda x: not self.InstanceCommand(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append(f'  PFN_{vkcommand.name} {vkcommand.name[2:]};\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
        out.append('};\n')

        out.append('''
void InitInstanceDispatchTable(VkInstance instance,
                               PFN_vkGetInstanceProcAddr pa,
                               InstanceDispatchTable *dt);
void InitDeviceDispatchTable(VkDevice device, PFN_vkGetDeviceProcAddr pa,
                             DeviceDispatchTable *dt);
''')
        self.write("".join(out))
        self.write(self.GenerateNamespaceEnd())

    def generateSource(self):
        self.write('#include "dispatch.h"')
        self.write(self.GenerateNamespaceBegin())

        out = []
        out.append('''
void InitInstanceDispatchTable(VkInstance instance,
                               PFN_vkGetInstanceProcAddr pa,
                               InstanceDispatchTable *dt) {
''')
        # The android loader complains if you look up these functions with a non-null instance handle
        global_commands = ('vkCreateInstance', 'vkEnumerateInstanceExtensionProperties')
        out.append(f'#ifdef VK_USE_PLATFORM_ANDROID_KHR\n')
        out.append(f'VkInstance global_instance_param = VK_NULL_HANDLE;\n')
        out.append(f'#else\n')
        out.append(f'VkInstance global_instance_param = instance;\n')
        out.append(f'#endif // VK_USE_PLATFORM_ANDROID_KHR\n')
        for vkcommand in filter(lambda x: self.InstanceCommand(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append(f'  dt->{vkcommand.name[2:]} =\n')
            instance_param = 'instance' if vkcommand.name not in global_commands else 'global_instance_param'
            out.append(f'    (PFN_{vkcommand.name})pa({instance_param}, "{vkcommand.name}");\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
        out.append('};\n\n')

        out.append('''void InitDeviceDispatchTable(VkDevice device, PFN_vkGetDeviceProcAddr pa,
                             DeviceDispatchTable *dt) {
''')
        for vkcommand in filter(lambda x: not self.InstanceCommand(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append(f'  dt->{vkcommand.name[2:]} =\n')
            out.append(f'    (PFN_{vkcommand.name})pa(device, "{vkcommand.name}");\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
        out.append('};\n\n')

        self.write("".join(out))
        self.write(self.GenerateNamespaceEnd())




#!/usr/bin/python3 -i
#
# Copyright (c) 2024-2025 Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from base_generator import BaseGenerator
from generators.generator_utils import PlatformGuardHelper

class TestIcdGenerator(BaseGenerator):
    def __init__(self):
        BaseGenerator.__init__(self)

        self.manual_functions = (
            'vkCreateInstance',
            'vkDestroyInstance',
            'vkAllocateCommandBuffers',
            'vkBeginCommandBuffer',
            'vkEndCommandBuffer',
            'vkResetCommandBuffer',
            'vkFreeCommandBuffers',
            'vkCreateCommandPool',
            'vkResetCommandPool',
            'vkDestroyCommandPool',
            'vkEnumeratePhysicalDevices',
            'vkCreateDevice',
            'vkDestroyDevice',
            'vkGetDeviceQueue',
            'vkGetDeviceQueue2',
            'vkEnumerateInstanceLayerProperties',
            'vkEnumerateInstanceVersion',
            'vkEnumerateDeviceLayerProperties',
            'vkEnumerateInstanceExtensionProperties',
            'vkEnumerateDeviceExtensionProperties',
            'vkGetPhysicalDeviceSurfacePresentModesKHR',
            'vkGetPhysicalDeviceSurfaceFormatsKHR',
            'vkGetPhysicalDeviceSurfaceFormats2KHR',
            'vkGetPhysicalDeviceSurfaceSupportKHR',
            'vkGetPhysicalDeviceSurfaceCapabilitiesKHR',
            'vkGetPhysicalDeviceSurfaceCapabilities2KHR',
            'vkGetInstanceProcAddr',
            'vkGetDeviceProcAddr',
            'vkGetPhysicalDeviceMemoryProperties',
            'vkGetPhysicalDeviceMemoryProperties2',
            'vkGetPhysicalDeviceQueueFamilyProperties',
            'vkGetPhysicalDeviceQueueFamilyProperties2',
            'vkGetPhysicalDeviceFeatures',
            'vkGetPhysicalDeviceFeatures2',
            'vkGetPhysicalDeviceFormatProperties',
            'vkGetPhysicalDeviceFormatProperties2',
            'vkGetPhysicalDeviceImageFormatProperties',
            'vkGetPhysicalDeviceImageFormatProperties2',
            'vkGetPhysicalDeviceSparseImageFormatProperties',
            'vkGetPhysicalDeviceSparseImageFormatProperties2',
            'vkGetPhysicalDeviceProperties',
            'vkGetPhysicalDeviceProperties2',
            'vkGetPhysicalDeviceExternalSemaphoreProperties',
            'vkGetPhysicalDeviceExternalFenceProperties',
            'vkGetPhysicalDeviceExternalBufferProperties',
            'vkGetBufferMemoryRequirements',
            'vkGetBufferMemoryRequirements2',
            'vkGetDeviceBufferMemoryRequirements',
            'vkGetImageMemoryRequirements',
            'vkGetImageMemoryRequirements2',
            'vkGetDeviceImageMemoryRequirements',
            'vkAllocateMemory',
            'vkFreeMemory',
            'vkMapMemory',
            'vkMapMemory2KHR',
            'vkUnmapMemory',
            'vkUnmapMemory2KHR',
            'vkBindBufferMemory',
            'vkBindBufferMemory2',
            'vkGetImageSubresourceLayout',
            'vkCreateSwapchainKHR',
            'vkDestroySwapchainKHR',
            'vkGetSwapchainImagesKHR',
            'vkAcquireNextImageKHR',
            'vkAcquireNextImage2KHR',
            'vkCreateBuffer',
            'vkDestroyBuffer',
            'vkCreateImage',
            'vkDestroyImage',
            'vkEnumeratePhysicalDeviceGroups',
            'vkGetPhysicalDeviceMultisamplePropertiesEXT',
            'vkGetPhysicalDeviceFragmentShadingRatesKHR',
            'vkGetPhysicalDeviceCalibrateableTimeDomainsKHR',
            'vkGetFenceWin32HandleKHR',
            'vkGetFenceFdKHR',
            'vkEnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR',
            'vkGetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR',
            'vkGetShaderModuleIdentifierEXT',
            'vkGetImageSparseMemoryRequirements',
            'vkGetImageSparseMemoryRequirements2',
            'vkGetBufferDeviceAddress',
            'vkGetDescriptorSetLayoutSizeEXT',
            'vkGetAccelerationStructureBuildSizesKHR',
            'vkGetAccelerationStructureMemoryRequirementsNV',
            'vkGetAccelerationStructureDeviceAddressKHR',
            'vkGetVideoSessionMemoryRequirementsKHR',
            'vkGetPhysicalDeviceCooperativeMatrixPropertiesKHR',
            'vkGetPhysicalDeviceVideoCapabilitiesKHR',
            'vkGetPhysicalDeviceVideoFormatPropertiesKHR',
            'vkGetDescriptorSetLayoutSupport',
            'vkGetRenderAreaGranularity',
            'vkGetMemoryFdKHR',
            'vkGetMemoryHostPointerPropertiesEXT',
            'vkGetAndroidHardwareBufferPropertiesANDROID',
            'vkGetPhysicalDeviceDisplayPropertiesKHR',
            'vkRegisterDisplayEventEXT',
            'vkGetMemoryWin32HandlePropertiesKHR',
            'vkRegisterDisplayEventEXT',
            'vkDeviceWaitIdle',
            'vkGetDeviceQueue',
            'vkQueueSubmit',
            'vkQueueSubmit2',
            'vkQueueBindSparse',
            'vkQueueWaitIdle',
            'vkCreateSemaphore',
            'vkDestroySemaphore',
            'vkGetSemaphoreCounterValue',
            'vkSignalSemaphore',
            'vkWaitSemaphores',
            'vkCreateFence',
            'vkDestroyFence',
            'vkResetFences',
            'vkWaitForFences',
            'vkGetFenceStatus',
            'vkGetDeviceFaultInfoEXT',
            'vkGetQueueCheckpointDataNV',
            'vkGetQueueCheckpointData2NV',
            'vkCmdBeginDebugUtilsLabelEXT',
            'vkCreatePipelineBinariesKHR',
        )

    def generate(self):
        self.write(f'''// *** THIS FILE IS GENERATED - DO NOT EDIT ***
            // See {os.path.basename(__file__)} for modifications

            /***************************************************************************
            *
            * Copyright (c) 2024-2025 LunarG, Inc.
            *
            * Licensed under the Apache License, Version 2.0 (the "License");
            * you may not use this file except in compliance with the License.
            * You may obtain a copy of the License at
            *
            *     http://www.apache.org/licenses/LICENSE-2.0
            *
            * Unless required by applicable law or agreed to in writing, software
            * distributed under the License is distributed on an "AS IS" BASIS,
            * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
            * See the License for the specific language governing permissions and
            * limitations under the License.
            ****************************************************************************/\n''')

        self.write('// NOLINTBEGIN') # Wrap for clang-tidy to ignore

        if self.filename == 'test_icd_helper.h':
            self.generateHeader()
        else:
            self.write(f'\nFile name {self.filename} has no code to generate\n')

        self.write('// NOLINTEND') # Wrap for clang-tidy to ignore

    def generateHeader(self):
        out = []
        out.append('''
            #pragma once

            #include <stdint.h>
            #include <cstring>
            #include <string>
            #include <unordered_map>
            #include <vulkan/vulkan.h>

            #include "test_icd_command.h"

            namespace icd {
            ''')
        guard_helper = PlatformGuardHelper()

        # Needed because things like VK_KHR_get_physical_device_properties2 profile layer checks in driver
        out.append('''
            // Map of instance extension name to version
            static const std::unordered_map<std::string, uint32_t> instance_extension_map = {
            ''')
        for extension in [x for x in self.vk.extensions.values() if x.instance]:
            out.extend(guard_helper.add_guard(extension.protect))
            out.append(f'{{{extension.nameString}, {extension.specVersion}}},\n')
        out.extend(guard_helper.add_guard(None))
        out.append('};\n')

        out.append('''
            // Map of device extension name to version
            static const std::unordered_map<std::string, uint32_t> device_extension_map = {
            ''')
        for extension in [x for x in self.vk.extensions.values() if not x.instance]:
            out.extend(guard_helper.add_guard(extension.protect))
            out.append(f'{{{extension.nameString}, {extension.specVersion}}},\n')
        out.extend(guard_helper.add_guard(None))
        out.append('};\n')

        for command in self.vk.commands.values():
            prototype = "static " + command.cPrototype
            prototype = prototype.replace("VKAPI_CALL vk", "VKAPI_CALL ")
            out.extend(guard_helper.add_guard(command.protect))
            out.append(f'{prototype}\n')
        out.extend(guard_helper.add_guard(None))

        out.append('''
            // Map of all APIs to be intercepted by this layer
            static const std::unordered_map<std::string, void*> name_to_func_ptr_map = {
            ''')

        for command in self.vk.commands.values():
            out.extend(guard_helper.add_guard(command.protect))
            out.append(f'{{"{command.name}", (void*){command.name[2:]}}},\n')
        out.extend(guard_helper.add_guard(None))
        out.append('};\n')

        for command in [x for x in self.vk.commands.values() if x.name not in self.manual_functions]:
            out.extend(guard_helper.add_guard(command.protect))
            prototype = "static " + command.cPrototype
            prototype = prototype.replace("VKAPI_CALL vk", "VKAPI_CALL ")
            prototype = prototype.replace(");", ") {")
            out.append(f'{prototype}\n')

            voidReturn = command.returnType == 'void'
            # For alias that are promoted, just point to new function
            if command.alias and command.alias in self.vk.commands:
                paramList = [param.name for param in command.params]
                params = ', '.join(paramList)
                returnName = '' if voidReturn else 'return '
                out.append(f'{returnName}{command.alias[2:]}({params});')
            elif command.name.startswith('vkCreate') or command.name.startswith('vkAllocate'):
                last_param = command.params[-1]
                out.append('unique_lock_t lock(global_lock);\n')
                if (last_param.length):
                    out.append(f'for (uint32_t i = 0; i < {last_param.length}; ++i) {{\n')
                    out.append(f'{last_param.name}[i] = ({last_param.type})global_unique_handle++;\n')
                    out.append('}\n')
                else:
                    out.append(f'*{last_param.name} = ({last_param.type})global_unique_handle++;\n')
                out.append('return VK_SUCCESS;\n')
            elif command.name.startswith('vkCmd'):
                paramList = [param.name for param in command.params]
                params = ', '.join(paramList)
                out.append('auto *cb = reinterpret_cast<CommandBuffer*>(commandBuffer);')
                name = command.name.replace('vk', '')

                out.append(f'cb->Tracker().{name}({params});')
                if not voidReturn:
                    out.append('return VK_SUCCESS;')
            elif not voidReturn:
                out.append('return VK_SUCCESS;')

            out.append('}\n\n')
        out.extend(guard_helper.add_guard(None))

        out.append('\n} // namespace icd')
        self.write("".join(out))



#!/usr/bin/python3 -i
#
# Copyright (c) 2023-2024 LunarG, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from generators.cdl_base_generator import CdlBaseOutputGenerator

class CommandTrackerGenerator(CdlBaseOutputGenerator):
    def __init__(self):
        CdlBaseOutputGenerator.__init__(self)

    #
    # Called at beginning of processing as file is opened
    def generate(self):
        file_start = self.GenerateFileStart(os.path.basename(__file__))
        self.write(file_start)

        if self.filename == 'command_tracker.h':
            self.generateCommandTrackerHeader()
        elif self.filename == 'command_tracker.cpp':
            self.generateCommandTrackerSource()
        else:
            self.write(f'\nFile name {self.filename} has no code to generate\n')

        self.write(self.GenerateFileEnd())

    def generateCommandTrackerHeader(self):
        out = []
        out.append('''

#include <string>
#include <vector>
#include <iostream>
#include <vulkan/vulkan.h>

#include "command_common.h"
#include "command_recorder.h"

class CommandTracker
{
 public:
  void Reset();

  const std::vector<Command> &GetCommands() const { return commands_; }
  std::vector<Command> &GetCommands() { return commands_; }

''')
        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            pre_func_decl = vkcommand.cPrototype.replace('VKAPI_ATTR ', '').replace('VKAPI_CALL ', '').replace(f'{vkcommand.returnType} ', 'void ', 1).replace('vk', '', 1)
            out.append(f'  {pre_func_decl}\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')
        out.append(''' private:
  std::vector<Command> commands_;
  CommandRecorder recorder_;
  std::vector<std::string> labels_;
};
''')
        self.write("".join(out))

    def generateCommandTrackerSource(self):
        out = []
        out.append('''
#include <vulkan/vulkan.h>
#include <cassert>

#include "command_common.h"
#include "command_tracker.h"

void CommandTracker::Reset()
{
  commands_.clear();
  recorder_.Reset();
}
''')


        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            func_decl = vkcommand.cPrototype.replace('VKAPI_ATTR ', '').replace('VKAPI_CALL ', '').replace(';', ' {').replace(f'{vkcommand.returnType} ', 'void ', 1).replace('vk', 'CommandTracker::', 1)
            out.append(f'{func_decl}\n')
            out.append('  Command cmd {};\n')
            out.append(f'  cmd.type = Command::Type::k{vkcommand.name[2:]};\n')
            out.append('  cmd.id = static_cast<uint32_t>(commands_.size()) + 1;\n')
            func_call = f'cmd.parameters = recorder_.Record{vkcommand.name[2:]}('

            if vkcommand.name == 'vkCmdDebugMarkerBeginEXT':
                out.append('  labels_.push_back(pMarkerInfo->pMarkerName);\n')
            elif vkcommand.name == 'vkCmdBeginDebugUtilsLabelEXT':
                out.append('  labels_.push_back(pLabelInfo->pLabelName);\n')
            out.append('  cmd.labels = labels_;\n')
            if vkcommand.name in ('vkCmdEndDebugUtilsLabelEXT', 'vkCmdDebugMarkerEndEXT'):
                out.append('  // do not crash even if the application ends without a marker present\n')
                out.append('  if (!labels_.empty()) {\n')
                out.append('      labels_.pop_back();\n')
                out.append('  }\n')

            count = 0
            for vkparam in vkcommand.params:
                if count != 0:
                    func_call += ', '
                func_call += vkparam.name
                count += 1
            out.append(f'  {func_call});\n')
            out.append('  commands_.push_back(cmd);\n')
            out.append('}\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')


        self.write("".join(out))



#!/usr/bin/python3 -i
#
# Copyright (c) 2023-2025 LunarG, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from generators.cdl_base_generator import CdlBaseOutputGenerator
#
# CommandPrinterOutputGenerator - Generate command print utilities
class CommandPrinterOutputGenerator(CdlBaseOutputGenerator):
    def __init__(self):
        CdlBaseOutputGenerator.__init__(self)

    #
    # Called at beginning of processing as file is opened
    def generate(self):
        file_start = self.GenerateFileStart(os.path.basename(__file__))
        self.write(file_start)

        if self.filename == 'command_printer.h':
            self.generateHeader()
        elif self.filename == 'command_printer_types.cpp':
            self.generateTypesSource()
        elif self.filename == 'command_printer_structs.cpp':
            self.generateStructsSource()
        elif self.filename == 'command_printer.cpp':
            self.generateSource()
        else:
            self.write(f'\nFile name {self.filename} has no code to generate\n')

        self.write(self.GenerateFileEnd())

    def generateHeader(self):
        self.write('\n#pragma once')

        out = []
        out.append('''
#include <yaml-cpp/emitter.h>
#include <vulkan/vulkan.h>

#include "command_common.h"
#include "object_name_db.h"

struct VkStruct {
    VkStructureType sType;
    void *pNext;
};

// Declare generic struct printer.
YAML::Emitter &PrintVkStruct(YAML::Emitter &os, const VkStruct *pStruct);

// Declare pNext chain printer.
YAML::Emitter &PrintNextPtr(YAML::Emitter &os, const void *pNext);
''')
        self.write("".join(out))

        self.write("// Declare Handle stream operators\n")
        out = []
        for vkhandle in self.vk.handles.values():
            out.extend([f'#ifdef {vkhandle.protect}\n'] if vkhandle.protect else [])
            if not vkhandle.dispatchable:
                out.append('#if VK_USE_64_BIT_PTR_DEFINES\n')
            out.append(f'YAML::Emitter &operator<<(YAML::Emitter& os, const {vkhandle.name} &a);')
            if not vkhandle.dispatchable:
                out.append('#endif //VK_USE_64_BIT_PTR_DEFINES\n')
            out.extend([f'#endif //{vkhandle.protect}\n'] if vkhandle.protect else [])
            out.append('\n')
        self.write("".join(out))
        self.write("\n// Declare stream operators for enums.\n")
        out = []
        for vkenum in [x for x in self.vk.enums.values() if len(x.fields) > 0]:
            out.extend([f'#ifdef {vkenum.protect}\n'] if vkenum.protect else [])
            out.append(f'YAML::Emitter &operator<<(YAML::Emitter &os, const {vkenum.name} &t);\n')
            out.extend([f'#endif //{vkenum.protect}\n'] if vkenum.protect else [])
        self.write("".join(out))

        self.write("\n// Declare all stream operators.\n")
        out = []
        for vkstruct in self.vk.structs.values():
            out.extend([f'#ifdef {vkstruct.protect}\n'] if vkstruct.protect else [])
            out.append(f'YAML::Emitter &operator<<(YAML::Emitter &os, const {vkstruct.name} &t);\n')
            out.extend([f'#endif //{vkstruct.protect}\n'] if vkstruct.protect else [])
        self.write("".join(out))

        out = []
        out.append('''
// Declare print functions.
class ObjectInfoDB;
class CommandPrinter {
 public:
  void SetNameResolver(const ObjectInfoDB *name_resolver);
  void PrintCommandParameters(YAML::Emitter &os, const Command &cmd);
''')
        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append(f'  void Print{vkcommand.name[2:]}Args(YAML::Emitter &os, const {vkcommand.name[2:]}Args &args);\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
        out.append('};\n');
        self.write("".join(out))

    def printArray(self, out, lengths, cur_length, prefix, member, outer = ''):
        variable = chr(ord('i') + cur_length)
        out.append('os << YAML::BeginSeq;\n')
        out.append(f'for (uint64_t {variable} = 0; {variable} < uint64_t({lengths[cur_length]}); ++{variable}) {{\n')
        if cur_length < (len(lengths) - 1):
            self.printArray(out, lengths, cur_length + 1, prefix, member, outer + '[' + variable + ']')
        else:
            pointer_prefix = prefix
            pointer_suffix = ''
            pointer_count = member.cDeclaration.count('*')
            if pointer_count > len(lengths):
                pointer_prefix = '*(' + prefix
                pointer_suffix = ')'
            out.append(f'os << {pointer_prefix}{member.name}{pointer_suffix}{outer}[{variable}];')
        out.append(f'}} // for {variable}\n')
        out.append('os << YAML::EndSeq;\n')

    def printMember(self, out, member, prefix, is_pointer):
        no_deref_structs = (
                'xcb_connection_t',
                'Display',
                'wl_display',
                'wl_surface',
                'ANativeWindow',
                'AHardwareBuffer',
                'SECURITY_ATTRIBUTES',
                'CAMetalLayer',
                'StdVideoEncodeH264SliceHeader',
                'StdVideoEncodeH264PictureInfo',
                'StdVideoEncodeH264ReferenceInfo',
                'StdVideoEncodeH265SliceSegmentHeader',
                'StdVideoEncodeH265PictureInfo',
                'StdVideoEncodeH265ReferenceInfo',
                'StdVideoDecodeH264PictureInfo',
                'StdVideoDecodeH264ReferenceInfo',
                'StdVideoDecodeH265PictureInfo',
                'StdVideoDecodeH265ReferenceInfo',
                'StdVideoAV1SequenceHeader',
                'StdVideoDecodeAV1PictureInfo',
                'StdVideoDecodeAV1ReferenceInfo',
                'StdVideoH264PictureParameterSet',
                'StdVideoH264SequenceParameterSet',
                'StdVideoH265PictureParameterSet',
                'StdVideoH265SequenceParameterSet',
                'StdVideoH265VideoParameterSet',
                'StdVideoEncodeAV1DecoderModelInfo',
                'StdVideoEncodeAV1OperatingPointInfo',
                'StdVideoEncodeAV1PictureInfo',
                'StdVideoEncodeAV1ReferenceInfo',
            )
        out.append(f'    os << YAML::Key <<  "{member.name}";\n')
        if is_pointer:
            prefix += '->'
        else:
            prefix += '.'
        if member.length is not None:

            lengths = []
            lengths_names = []
            if len(member.fixedSizeArray) > 1 or member.length.count(',') > 0:
                lengths = member.length.split(',')
                if len(lengths) > 1 and lengths[len(lengths) - 1] == '1':
                    lengths.pop()
            else:
                lengths.append(member.length)
            # for truly static arrays these are equal
            static_array = (lengths == member.fixedSizeArray)

            array_type = ''
            if member.const:
                array_type = 'Const'
            for i in range(len(lengths)):
                length_name = ''
                if '*' in lengths[i] or '+' in lengths[i] or '/' in lengths[i]:
                    elements = lengths[i].split(' ')
                    for j in range(len(elements)):
                        for k in range(len(elements[j])):
                            if elements[j][k] == '(':
                                continue
                            if elements[j][k].islower():
                                if k == 0:
                                    temp = prefix + elements[j]
                                else:
                                    temp = elements[j][:k] + prefix + elements[j][k:]
                                elements[j] = temp
                                break
                    lengths[i] = ''.join(elements)
                    length_name = '('
                    length_name += lengths[i] + ')'
                elif i == 0 and not static_array:
                    length_name = prefix + lengths[i]
                else:
                    length_name = lengths[i]
                if static_array:
                    array_type += 'Fixed'
                else:
                    array_type += 'Dynamic'
                array_type += 'Array('
                if i == len(lengths) - 1:
                    array_type += member.type
                lengths_names.append(length_name)
            for i in range(len(lengths)):
                array_type += ')'

            if member.type == 'char' and len(lengths) == 1:
                out.append(f'    // {member.name} -> Field -> {array_type}\n')
                out.append(f'    os << YAML::Value << {prefix}{member.name};\n')
            elif member.type == 'void':
                if static_array:
                    out.append('    {\n')
                else:
                    out.append(f'    if ({lengths_names[0]} == 0) {{\n')
                    out.append('      os << YAML::Value << "nullptr";\n')
                    out.append('    } else {\n')
                out.append('      os << YAML::Value << YAML::BeginSeq;\n')
                out.append('      {\n')
                out.append(f'        const uint8_t *p = (const uint8_t *){prefix}{member.name};\n')
                out.append(f'        for (uint64_t i = 0; i < {lengths_names[0]}; ++i) {{\n')
                out.append('          os << crash_diagnostic_layer::Uint8ToStr(p[i]);\n')
                out.append('        }\n')
                out.append('      }\n')
                out.append('      os << YAML::EndSeq;\n')
                out.append('    }\n')
            else:
                out.append(f'    // {member.name} -> Field -> {array_type}\n')
                if static_array:
                    out.append('    {\n')
                else:
                    out.append(f'    if ({lengths_names[0]} == 0) {{\n')
                    out.append('      os << YAML::Value << "nullptr";\n')
                    out.append('    } else {\n')
                out.append('      os << YAML::Value;\n')
                out.append('      {\n')
                out.append(f'       os << YAML::Comment("{member.type}");\n')
                if member.type not in no_deref_structs:
                    if member.type in self.vk.structs:
                        if member.type not in no_deref_structs:
                            self.printArray(out, lengths_names, 0, prefix, member)
                        else:
                            out.append(f'       os << YAML::BeginSeq << "{member.type} not decoded" << YAML::EndSeq;\n')
                    else:
                        self.printArray(out, lengths_names, 0, prefix, member)
                out.append('      }\n')
                out.append('    }\n')
        elif member.pointer:
            if member.name == 'pNext':
                out.append('    // pNext -> Field -> ConstNextPtr(void)\n')
                out.append('    os << YAML::Value << YAML::BeginSeq;\n')
                out.append(f'    PrintNextPtr(os, {prefix}pNext);\n')
                out.append('    os << YAML::EndSeq;\n')
            elif member.type == 'void':
                out.append('    // void\n')
                out.append('    os << YAML::Value << "NOT_AVAILABLE";\n')
            elif member.type == 'char':
                out.append(f'    // {member.name} -> Field -> string\n')
                out.append(f'    os << YAML::Value << {prefix}{member.name};\n')
            elif member.type in no_deref_structs:
                out.append('    // non-dereferenced pointer\n')
                out.append(f'   os << YAML::Value << {prefix}{member.name};\n')
            else:
                out.append('    // pointer\n')
                out.append(f'    if ({prefix}{member.name} != nullptr) {{\n')
                if member.type in self.vk.structs:
                    out.append(f'      os << YAML::Value << *{prefix}{member.name};\n')
                else:
                    out.append(f'      os << YAML::Value << *{prefix}{member.name};\n')
                out.append('    } else {\n')
                out.append('      os << YAML::Value << "nullptr";\n')
                out.append('    }\n')
        else:
            out.append(f'    // {member.name} -> Field -> {member.type}\n')
            if member.type == 'VkDeviceAddress':
                out.append(f'    os << YAML::Value << crash_diagnostic_layer::Uint64ToStr({prefix}{member.name});\n')
            else:
                out.append(f'    os << YAML::Value << {prefix}{member.name};\n')

    def generateTypesSource(self):
        out = []
        out.append('''
#include <streambuf>
#include <vulkan/vk_enum_string_helper.h>

#include "command_common.h"
#include "command_printer.h"
#include "util.h"

const ObjectInfoDB *global_name_resolver = nullptr;
void CommandPrinter::SetNameResolver(const ObjectInfoDB *name_resolver) {
  global_name_resolver = name_resolver;
}
''')
        self.write("".join(out))

        self.write("// Handle stream operators\n")
        out = []
        for vkhandle in self.vk.handles.values():
            out.extend([f'#ifdef {vkhandle.protect}\n'] if vkhandle.protect else [])
            if not vkhandle.dispatchable:
                out.append('#if VK_USE_64_BIT_PTR_DEFINES\n')
            out.append(f'''
YAML::Emitter &operator<<(YAML::Emitter& os, const {vkhandle.name} &a) {{
    os << global_name_resolver->GetObjectInfo(reinterpret_cast<uint64_t>(a));
    return os;
}}
''')
            if not vkhandle.dispatchable:
                out.append('#endif //VK_USE_64_BIT_PTR_DEFINES\n')
            out.extend([f'#endif //{vkhandle.protect}\n'] if vkhandle.protect else [])
            out.append('\n')
        self.write("".join(out))

        self.write("\n// Define stream operators for enums.\n")
        out = []
        for vkenum in [x for x in self.vk.enums.values() if len(x.fields) > 0]:
            out.extend([f'#ifdef {vkenum.protect}\n'] if vkenum.protect else [])
            out.append(f'YAML::Emitter &operator<<(YAML::Emitter & os, const {vkenum.name} &t) {{\n')
            out.append(f'  os << string_{vkenum.name}(t);\n')
            out.append('  return os;\n')
            out.append('}\n')
            out.extend([f'#endif //{vkenum.protect}\n'] if vkenum.protect else [])
            out.append('\n')
        self.write("".join(out))


    def generateStructsSource(self):
        manual_structs = ('VkWriteDescriptorSet', 'VkAccelerationStructureBuildGeometryInfoKHR')
        out = []
        out.append('''
#include <streambuf>
#include <vulkan/vk_enum_string_helper.h>

#include "command_common.h"
#include "command_printer.h"
#include "util.h"

YAML::Emitter &PrintNextPtr(YAML::Emitter &os, const void *pNext) {
  if (pNext == nullptr) {
    os << "nullptr";
    return os;
  }

  const VkStruct *pStruct = reinterpret_cast<const VkStruct *>(pNext);
  PrintVkStruct(os, pStruct);
  return PrintNextPtr(os, pStruct->pNext);
}
''')
        self.write("".join(out))
        self.write("\n// Define all stream operators.\n")
        out = []
        for vkstruct in self.vk.structs.values():
            if vkstruct.name in manual_structs:
                continue
            out.extend([f'#ifdef {vkstruct.protect}\n'] if vkstruct.protect else [])
            out.append(f'YAML::Emitter &operator<<(YAML::Emitter & os, const {vkstruct.name} &t) {{\n')
            out.append('  os << YAML::BeginMap;\n')
            for member in vkstruct.members:
                self.printMember(out, member, 't', False)
            out.append('  os << YAML::EndMap;\n')
            out.append('  return os;\n')
            out.append('}\n')
            out.extend([f'#endif //{vkstruct.protect}\n'] if vkstruct.protect else [])
            out.append('\n')
        self.write("".join(out))

        out = []
        out.append('''
YAML::Emitter &operator<<(YAML::Emitter &os, const VkWriteDescriptorSet &t) {
  os << YAML::BeginMap;
  os << YAML::Key << "sType" << YAML::Value << t.sType;

  os << YAML::Key << "pNext" << YAML::Value << t.pNext;

  os << YAML::Key << "dstSet" << YAML::Value << t.dstSet;

  os << YAML::Key << "dstBinding" << YAML::Value << t.dstBinding;

  os << YAML::Key << "dstArrayElement" << YAML::Value << t.dstArrayElement;

  os << YAML::Key << "descriptorCount" << YAML::Value << t.descriptorCount;

  os << YAML::Key << "descriptorType" << YAML::Value << t.descriptorType;

  switch (t.descriptorType){
    case VK_DESCRIPTOR_TYPE_SAMPLER:
    case VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER:
    case VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE:
    case VK_DESCRIPTOR_TYPE_STORAGE_IMAGE:
      os << YAML::Key << "pImageInfo" << YAML::Value << YAML::BeginSeq;
      for (uint32_t i = 0; i < t.descriptorCount; ++i) {
        os << t.pImageInfo[i];
      }
      os << YAML::EndSeq;
      break;

    case VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER:
    case VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER:
      os << YAML::Key << "pTexelBufferView" << YAML::Value << YAML::BeginSeq;
      for (uint32_t i = 0; i < t.descriptorCount; ++i) {
        os << t.pTexelBufferView[i];
      }
      os << YAML::EndSeq;
      break;

    case VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER:
    case VK_DESCRIPTOR_TYPE_STORAGE_BUFFER:
    case VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC:
    case VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC:
    case VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT:
      os << YAML::Key << "pBufferInfo" << YAML::Value << YAML::BeginSeq;
      for (uint32_t i = 0; i < t.descriptorCount; ++i) {
        os << t.pBufferInfo[i];
      }
      os << YAML::EndSeq;
      break;

    default:
      os << YAML::Key << "Unknown Descriptor Type" << YAML::Value << t.descriptorType;
  }
  os << YAML::EndMap;
  return os;
}

YAML::Emitter &operator<<(YAML::Emitter &os, const VkAccelerationStructureBuildGeometryInfoKHR &t) {
    os << YAML::BeginMap;
    os << YAML::Key << "sType";
    // sType -> Field -> VkStructureType
    os << YAML::Value << t.sType;
    os << YAML::Key << "pNext";
    // pNext -> Field -> ConstNextPtr(void)
    os << YAML::Value << YAML::BeginSeq;
    PrintNextPtr(os, t.pNext);
    os << YAML::EndSeq;
    os << YAML::Key << "type";
    // type -> Field -> VkAccelerationStructureTypeKHR
    os << YAML::Value << t.type;
    os << YAML::Key << "flags";
    // flags -> Field -> VkBuildAccelerationStructureFlagsKHR
    os << YAML::Value << t.flags;
    os << YAML::Key << "mode";
    // mode -> Field -> VkBuildAccelerationStructureModeKHR
    os << YAML::Value << t.mode;
    os << YAML::Key << "srcAccelerationStructure";
    // srcAccelerationStructure -> Field -> VkAccelerationStructureKHR
    os << YAML::Value << t.srcAccelerationStructure;
    os << YAML::Key << "dstAccelerationStructure";
    // dstAccelerationStructure -> Field -> VkAccelerationStructureKHR
    os << YAML::Value << t.dstAccelerationStructure;
    os << YAML::Key << "geometryCount";
    // geometryCount -> Field -> uint32_t
    os << YAML::Value << t.geometryCount;
    // This part needs to be manual because we don't have an easy way to automatically handle
    // the "geometryCount refers to either pGeometries or ppGeometries" logic.
    if (t.pGeometries) {
        os << YAML::Key << "pGeometries";
        // pGeometries -> Field -> ConstDynamicArray(VkAccelerationStructureGeometryKHR)
        if (t.geometryCount == 0) {
            os << YAML::Value << "nullptr";
        } else {
            os << YAML::Value;
            {
                os << YAML::Comment("VkAccelerationStructureGeometryKHR");
                os << YAML::BeginSeq;
                for (uint64_t i = 0; i < uint64_t(t.geometryCount); ++i) {
                    os << t.pGeometries[i];
                }  // for i
                os << YAML::EndSeq;
            }
        }
    } else if (t.ppGeometries) {
        os << YAML::Key << "ppGeometries";
        // ppGeometries -> Field -> ConstDynamicArray(VkAccelerationStructureGeometryKHR)
        if (t.geometryCount == 0) {
            os << YAML::Value << "nullptr";
        } else {
            os << YAML::Value;
            {
                os << YAML::Comment("VkAccelerationStructureGeometryKHR");
                os << YAML::BeginSeq;
                for (uint64_t i = 0; i < uint64_t(t.geometryCount); ++i) {
                    os << *(t.ppGeometries)[i];
                }  // for i
                os << YAML::EndSeq;
            }
        }
    }
    os << YAML::Key << "scratchData";
    // scratchData -> Field -> VkDeviceOrHostAddressKHR
    os << YAML::Value << t.scratchData;
    os << YAML::EndMap;
    return os;
}
''')
        self.write("".join(out))

        out = []
        out.append('//  Print out a VkStruct\n')
        out.append('YAML::Emitter & PrintVkStruct(YAML::Emitter & os, const VkStruct *pStruct) {\n')
        out.append('  switch (pStruct->sType) {\n')
        for vkstruct in self.vk.structs.values():
            if vkstruct.sType is None:
                continue
            out.extend([f'#ifdef {vkstruct.protect}\n'] if vkstruct.protect else [])
            out.append(f'  case {vkstruct.sType}:\n')
            out.append(f'    os << *reinterpret_cast<const {vkstruct.name} *>(pStruct);\n')
            out.append('    break;\n')
            out.extend([f'#endif //{vkstruct.protect}\n'] if vkstruct.protect else [])

        out.append('  default: break;\n')
        out.append('  }\n')
        out.append('  return os;\n')
        out.append('}\n')
        self.write("".join(out))

    def generateSource(self):
        out = []
        out.append('''
#include <streambuf>
#include <vulkan/vk_enum_string_helper.h>

#include "command_common.h"
#include "command_printer.h"
#include "util.h"

''')
        self.write("".join(out))
        self.write("\n// Define print functions.\n")
        out = []
        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append(f'void CommandPrinter::Print{vkcommand.name[2:]}Args(\n')
            out.append(f'  YAML::Emitter & os, const {vkcommand.name[2:]}Args &args) {{\n')
            for member in vkcommand.params:
                if member.name != 'commandBuffer':
                    self.printMember(out, member, 'args', False)
            out.append('}\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')
        out.append('''
void CommandPrinter::PrintCommandParameters(YAML::Emitter &os, const Command &cmd)
{
  switch (cmd.type)
  {
    default:
    case Command::Type::kUnknown:
      // output an empty map for consistency with other command printers
      os << YAML::BeginMap << YAML::EndMap;
      break;
''')
        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append(f'    case Command::Type::k{vkcommand.name[2:]}:\n')
            out.append('      if (cmd.parameters) {\n')
            out.append(f'        auto args = reinterpret_cast<{vkcommand.name[2:]}Args *>(cmd.parameters);\n')
            out.append(f'        Print{vkcommand.name[2:]}Args(os, *args);\n')
            out.append('      }\n')
            out.append('      break;\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')
        out.append('    } // switch (cmd.type)\n\n')
        out.append('} // CommandPrinter::PrintCommandParameters\n\n')
        self.write("".join(out))



#!/usr/bin/python3 -i
#
# Copyright (c) 2023-2024 LunarG, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from generators.cdl_base_generator import CdlBaseOutputGenerator

#NOTE: This generator only handles command buffer related overrides.
# Everything else should be done manually in cdl.h/cpp.
custom_pre_intercept_functions = (
   'vkBeginCommandBuffer',
   'vkResetCommandBuffer',
   'vkCmdBindPipeline'
)

custom_post_intercept_functions = ()

#
# InterceptCommandsOutputGenerator - Generate the dispatch tables
class InterceptContextGenerator(CdlBaseOutputGenerator):
    def __init__(self):
        CdlBaseOutputGenerator.__init__(self)

    #
    # Called at beginning of processing as file is opened
    def generate(self):
        file_start = self.GenerateFileStart(os.path.basename(__file__))
        self.write(file_start)

        if self.filename == 'cdl_commands.h.inc':
            self.generateContextCommandsHeader()
        elif self.filename == 'cdl_commands.cpp.inc':
            self.generateContextCommandsSource()
        else:
            self.write(f'\nFile name {self.filename} has no code to generate\n')

        self.write(self.GenerateFileEnd())

    def generateContextCommandsHeader(self):
        out = []
        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            post_func_decl = vkcommand.cPrototype.replace('VKAPI_ATTR ', '').replace('VKAPI_CALL ', '').replace(' vk', ' Post', 1).replace(';', ' override;')
            pre_func_decl = post_func_decl.replace('Post', 'Pre', 1)
            override_func_decl = post_func_decl.replace('Post', '', 1)
            if self.CommandHasReturn(vkcommand):
                post_func_decl = post_func_decl.replace(')', f',\n    {vkcommand.returnType}                                    result)')
            if self.InterceptPreCommand(vkcommand):
                out.append(f'{pre_func_decl}\n')
            if self.InterceptPostCommand(vkcommand):
                out.append(f'{post_func_decl}\n')
            if self.InterceptOverrideCommand(vkcommand):
                out.append(f'{override_func_decl}\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')
        self.write("".join(out))

    def generateContextCommandsSource(self):
        out = []

        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            post_func_decl = vkcommand.cPrototype.replace('VKAPI_ATTR ', '').replace('VKAPI_CALL ', '').replace(';', ' {').replace(' vk', ' Context::Post', 1)
            pre_func_decl = post_func_decl.replace('Context::Post', 'Context::Pre', 1)
            pre_func_call = '  '
            post_func_call = '  '
            if self.CommandHasReturn(vkcommand):
                post_func_decl = post_func_decl.replace(')', f',\n    {vkcommand.returnType}                                    result)')
                pre_func_call += 'return '
                post_func_call += 'return '
            pre_func_call += f'p_cmd->Pre{vkcommand.name[2:]}('
            post_func_call += f'p_cmd->Post{vkcommand.name[2:]}('
            count = 0
            for vkparam in vkcommand.params:
                if count != 0:
                    pre_func_call += ', '
                    post_func_call += ', '
                pre_func_call += vkparam.name
                post_func_call += vkparam.name
                count += 1
            if self.CommandHasReturn(vkcommand):
                post_func_call += ', result'
            pre_func_call += ');'
            post_func_call += ');'

            # Skip vkCmdBindPipeline pre call since it's implemented by hand
            if not (vkcommand.name in custom_pre_intercept_functions):
                out.append(f'{pre_func_decl}\n')
                out.append('  auto p_cmd = crash_diagnostic_layer::GetCommandBuffer(commandBuffer);\n')
                out.append(f'{pre_func_call}\n')
                out.append('}\n')

            if not (vkcommand.name in custom_post_intercept_functions):
                out.append(f'{post_func_decl}\n')
                out.append('  auto p_cmd = crash_diagnostic_layer::GetCommandBuffer(commandBuffer);\n')
                out.append(f'{post_func_call}\n')
                out.append('}\n')

            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')
        self.write("".join(out))




#!/usr/bin/env python3
# Copyright (c) 2023-2024 Valve Corporation
# Copyright (c) 2023-2024 LunarG, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTE: Android this documentation is crucial for understanding the layout of the NDK.
# https://android.googlesource.com/platform/ndk/+/master/docs/BuildSystemMaintainers.md

# NOTE: Environment variables we can rely on users/environments setting.
# https://github.com/actions/runner-images/blob/main/images/linux/Ubuntu2204-Readme.md#environment-variables-2

import argparse
import os
import sys
import shutil
import common_ci

# Manifest file describing out test application
def get_android_manifest() -> str:
    manifest = common_ci.RepoRelative('tests/android/AndroidManifest.xml')
    if not os.path.isfile(manifest):
        print(f"Unable to find manifest for APK! {manifest}")
        sys.exit(-1)
    return manifest

# Generate the APK from the CMake binaries
def generate_apk(SDK_ROOT : str, CMAKE_INSTALL_DIR : str) -> str:
    apk_dir = common_ci.RepoRelative('build-android/bin')

    # Delete APK directory since it could contain files from old runs
    if os.path.isdir(apk_dir):
        shutil.rmtree(apk_dir)

    shutil.copytree(CMAKE_INSTALL_DIR, apk_dir)

    android_manifest = get_android_manifest()

    # 33 is matching the targetSdkVersion
    android_jar = f"{SDK_ROOT}/platforms/android-33/android.jar"
    if not os.path.isfile(android_jar):
        print(f"Unable to find {android_jar}!")
        sys.exit(-1)

    apk_name = 'CrashDiagnosticLayerTests'

    unaligned_apk = f'{apk_dir}/{apk_name}-unaligned.apk'
    test_apk = f'{apk_dir}/{apk_name}.apk'

    # Create APK
    common_ci.RunShellCmd(f'aapt package -f -M {android_manifest} -I {android_jar} -F {unaligned_apk} {CMAKE_INSTALL_DIR}')

    # Align APK
    common_ci.RunShellCmd(f'zipalign -f 4 {unaligned_apk} {test_apk}')

    # Create Key (If it doesn't already exist)
    debug_key = common_ci.RepoRelative(f'{apk_dir}/debug.keystore')
    ks_pass = 'android'
    if not os.path.isfile(debug_key):
        dname = 'CN=Android-Debug,O=Android,C=US'
        common_ci.RunShellCmd(f'keytool -genkey -v -keystore {debug_key} -alias androiddebugkey -storepass {ks_pass} -keypass {ks_pass} -keyalg RSA -keysize 2048 -validity 10000 -dname {dname}')

    # Sign APK
    common_ci.RunShellCmd(f'apksigner sign --verbose --ks {debug_key} --ks-pass pass:{ks_pass} {test_apk}')

# Android APKs can contain binaries for multiple ABIs (armeabi-v7a, arm64-v8a, x86, x86_64).
# https://en.wikipedia.org/wiki/Apk_(file_format)#Package_contents
#
# As a result CMake will need to be run multiple times to create a complete test APK that can be run on any Android device.
def main():
    configs = ['Release', 'Debug', 'MinSizeRel']

    parser = argparse.ArgumentParser()
    parser.add_argument('--config', type=str, choices=configs, default=configs[0])
    parser.add_argument('--app-abi', dest='android_abi', type=str, default="arm64-v8a")
    parser.add_argument('--app-stl', dest='android_stl', type=str, choices=["c++_static", "c++_shared"], default="c++_static")
    parser.add_argument('--tests', action='store_true', help='Build unit tests')
    parser.add_argument('--apk', action='store_true', help='Generate an APK as a post build step.')
    parser.add_argument('--clean', action='store_true', help='Cleans CMake build artifacts')
    args = parser.parse_args()

    cmake_config = args.config
    android_abis = args.android_abi.split(" ")
    android_stl = args.android_stl
    build_tests = args.tests or args.apk
    create_apk = args.apk
    clean = args.clean

    if "ANDROID_NDK_HOME" not in os.environ:
        print("Cannot find ANDROID_NDK_HOME!")
        sys.exit(1)

    android_ndk_home = os.environ.get('ANDROID_NDK_HOME')
    android_toolchain = f'{android_ndk_home}/build/cmake/android.toolchain.cmake'

    # The only tool we require for building is CMake/Ninja
    required_cli_tools = ['cmake', 'ninja']

    # If we are building an APK we need a few more tools.
    if create_apk:
        if "ANDROID_SDK_ROOT" not in os.environ:
            print("Cannot find ANDROID_SDK_ROOT!")
            sys.exit(1)

        android_sdk_root = os.environ.get('ANDROID_SDK_ROOT')
        print(f"ANDROID_SDK_ROOT = {android_sdk_root}")
        required_cli_tools += ['aapt', 'zipalign', 'keytool', 'apksigner']

    print(f"ANDROID_NDK_HOME = {android_ndk_home}")
    print(f"Build configured for {cmake_config} | {android_stl} | {android_abis} | APK {create_apk}")

    if not os.path.isfile(android_toolchain):
        print(f'Unable to find android.toolchain.cmake at {android_toolchain}')
        exit(-1)

    for tool in required_cli_tools:
        path = shutil.which(tool)
        if path is None:
            print(f"Unable to find {tool}!")
            exit(-1)

        print(f"Using {tool} : {path}")

    cmake_install_dir = common_ci.RepoRelative('build-android/libs')

    # Delete install directory since it could contain files from old runs
    if os.path.isdir(cmake_install_dir):
        print("Cleaning CMake install")
        shutil.rmtree(cmake_install_dir)

    for abi in android_abis:
        build_dir = common_ci.RepoRelative(f'build-android/cmake/{abi}')
        lib_dir = f'lib/{abi}'

        if clean:
            print("Deleting CMakeCache.txt")

            # Delete CMakeCache.txt to ensure clean builds
            # NOTE: CMake 3.24 has --fresh which would be better to use in the future.
            cmake_cache = f'{build_dir}/CMakeCache.txt'
            if os.path.isfile(cmake_cache):
                os.remove(cmake_cache)

        cmake_cmd =  f'cmake -S . -B {build_dir} -G Ninja'

        cmake_cmd += f' -D CMAKE_BUILD_TYPE={cmake_config}'
        cmake_cmd += f' -D UPDATE_DEPS=ON -D UPDATE_DEPS_DIR={build_dir}'
        cmake_cmd += f' -D CMAKE_TOOLCHAIN_FILE={android_toolchain}'
        cmake_cmd += f' -D CMAKE_ANDROID_ARCH_ABI={abi}'
        cmake_cmd += f' -D CMAKE_INSTALL_LIBDIR={lib_dir}'
        cmake_cmd += f' -D BUILD_TESTS={build_tests}'
        cmake_cmd += f' -D CMAKE_ANDROID_STL_TYPE={android_stl}'

        cmake_cmd += ' -D ANDROID_PLATFORM=26'
        cmake_cmd += ' -D ANDROID_USE_LEGACY_TOOLCHAIN_FILE=NO'

        common_ci.RunShellCmd(cmake_cmd)

        build_cmd = f'cmake --build {build_dir}'
        common_ci.RunShellCmd(build_cmd)

        install_cmd = f'cmake --install {build_dir} --prefix {cmake_install_dir}'
        if cmake_config in ['Release', 'MinSizeRel']:
            install_cmd += f' --strip'
        common_ci.RunShellCmd(install_cmd)

    if create_apk:
        generate_apk(SDK_ROOT = android_sdk_root, CMAKE_INSTALL_DIR = cmake_install_dir)

if __name__ == '__main__':
    main()



#!/usr/bin/python3 -i
#
# Copyright 2023-2024 LunarG, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from generators.cdl_base_generator import CdlBaseOutputGenerator, InterceptFlag

#
# LayerBaseOutputGenerator - Generate command print utilities
class LayerBaseOutputGenerator(CdlBaseOutputGenerator):
    def __init__(self):
        CdlBaseOutputGenerator.__init__(self)

    #
    # Called at beginning of processing as file is opened
    def generate(self):
        file_start = self.GenerateFileStart(os.path.basename(__file__))
        self.write(file_start)
        if self.filename == 'layer_base.h.inc':
            self.generateHeader()
        elif self.filename == 'layer_base.cpp.inc':
            self.generateSource()
        else:
            self.write(f'\nFile name {self.filename} has no code to generate\n')
        self.write(self.GenerateFileEnd())

    def generateHeader(self):
        self.write('\n#pragma once')

        out = []
        for vkcommand in self.vk.commands.values():
            attrs = self.intercept_functions.get(vkcommand.name, None)
            if attrs is None:
                if self.CommandBufferCall(vkcommand):
                    attrs = InterceptFlag.NORMAL
                else:
                    continue
            if attrs & InterceptFlag.PRE:
                out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
                func_call = '    virtual ' + vkcommand.cPrototype.replace('\n','\n    ').replace('VKAPI_ATTR ', '').replace('VKAPI_CALL ', '')
                func_call = func_call.replace(' vk', ' Pre')
                if vkcommand.returnType is not None and vkcommand.returnType != 'void':
                    func_call = func_call.replace(';', ' { return VK_SUCCESS; }')
                else:
                    func_call = func_call.replace(';', ' {}')
                out.append(f'{func_call}\n')
                out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
                out.append('\n')
            if attrs & InterceptFlag.POST:
                out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
                func_call = '    virtual ' + vkcommand.cPrototype.replace('\n', '\n    ').replace('VKAPI_ATTR ', '').replace('VKAPI_CALL ', '')
                func_call = func_call.replace(' vk', ' Post')

                if vkcommand.returnType is not None and vkcommand.returnType != 'void':
                    func_call = func_call.replace(');', f',\n        {vkcommand.returnType}                                    result);')
                if vkcommand.returnType is not None and vkcommand.returnType != 'void':
                    func_call = func_call.replace(';', ' { return result; }')
                else:
                    func_call = func_call.replace(';', ' {}')
                out.append(f'{func_call}\n')
                out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
                out.append('\n')
            if attrs & InterceptFlag.OVERRIDE:
                out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
                func_call = '    virtual ' + vkcommand.cPrototype.replace('\n', '\n    ').replace('VKAPI_ATTR ', '').replace('VKAPI_CALL ', '')
                func_call = func_call.replace(' vk', ' ')
                if vkcommand.returnType is not None and vkcommand.returnType != 'void':
                    func_call = func_call.replace(';', ' = 0;')
                out.append(f'{func_call}\n')
                out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
                out.append('\n')
        self.write("".join(out))

        out = []

    def getApiFunctionType(self, command):
            if command.name in [
                    'vkCreateInstance',
                    'vkEnumerateInstanceVersion',
                    'vkEnumerateInstanceLayerProperties',
                    'vkEnumerateInstanceExtensionProperties',
                ]:
                return 'kFuncTypeInst'
            elif command.params[0].type == 'VkInstance':
                return'kFuncTypeInst'
            elif command.params[0].type == 'VkPhysicalDevice':
                return'kFuncTypePdev'
            else:
                return'kFuncTypeDev'

    def generateSource(self):
        out = []
        for vkcommand in self.vk.commands.values():
            attrs = self.intercept_functions.get(vkcommand.name, None)
            if attrs is None:
                if self.CommandBufferCall(vkcommand):
                    attrs = InterceptFlag.NORMAL
                else:
                    continue
            if attrs & (InterceptFlag.CUSTOM | InterceptFlag.OVERRIDE):
                continue

            lower_function_call = vkcommand.name + '('
            count = 0
            for vkparam in vkcommand.params:
                if count > 0:
                    lower_function_call += ', '
                lower_function_call += vkparam.name
                count += 1
            lower_function_call += ');\n'

            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            func_call = 'static ' + vkcommand.cPrototype.replace(' vk', ' Intercept').replace(';', ' {')
            out.append(f'{func_call}\n')

            if self.CommandHasReturn(vkcommand):
                default_value = self.GetDefaultReturnValue(vkcommand.returnType)
                out.append(f'  {vkcommand.returnType} result = {default_value};\n')
                out.append('\n')

            if self.InstanceCommand(vkcommand):
                out.append(f'  auto layer_data = GetInstanceLayerData(DataKey({vkcommand.params[0].name}));\n')
            else:
                out.append(f'  auto layer_data = GetDeviceLayerData(DataKey({vkcommand.params[0].name}));\n')

            if attrs & InterceptFlag.PRE:
              pre_func = lower_function_call.replace('vk', 'layer_data->interceptor->Pre')
              out.append(f'  {pre_func}')
              out.append('\n')

            out.append(f'  PFN_{vkcommand.name} pfn = layer_data->dispatch_table.{vkcommand.name[2:]};\n')
            out.append('  if (pfn != nullptr) {\n')
            out.append('    ')
            if self.CommandHasReturn(vkcommand):
                out.append('result = ')
            pfn_call = lower_function_call.replace(f'{vkcommand.name}', 'pfn')
            out.append(f'{pfn_call}')
            out.append('  }\n')
            out.append('\n')

            if attrs & InterceptFlag.POST:
              out.append('  ')
              post_call = lower_function_call.replace('vk', 'layer_data->interceptor->Post')
              if self.CommandHasReturn(vkcommand):
                  out.append('result = ')
                  post_call = post_call.replace(')', ', result)')
              out.append(f'{post_call}')

            if self.CommandHasReturn(vkcommand):
                out.append('  return result;\n')
            out.append('}\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')
        self.write("".join(out))

        out = []
        out.append('''
typedef enum ApiFunctionType { kFuncTypeInst = 0, kFuncTypePdev = 1, kFuncTypeDev = 2 } ApiFunctionType;

typedef struct {
    ApiFunctionType function_type;
    void* funcptr;
} function_data;

const std::unordered_map<std::string, function_data> &GetNameToFuncPtrMap() {
    static const std::unordered_map<std::string, function_data> name_to_func_ptr_map = {
''')
        for command in filter(self.NeedsIntercept, self.vk.commands.values()):
            out.extend([f'#ifdef {command.protect}\n'] if command.protect else [])
            out.append(f'    {{"{command.name}", {{{self.getApiFunctionType(command)}, (void*)Intercept{command.name[2:]}}}}},\n')
            out.extend([f'#endif //{command.protect}\n'] if command.protect else [])
        out.append('};\n')
        out.append(' return name_to_func_ptr_map;\n')
        out.append('};\n')
        self.write("".join(out))




#!/usr/bin/python3 -i
#
# Copyright (c) 2023-2025 LunarG, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from generators.cdl_base_generator import CdlBaseOutputGenerator

#
# CommandRecorderOutputGenerator - Generate the dispatch tables
class CommandRecorderOutputGenerator(CdlBaseOutputGenerator):
    def __init__(self):
        CdlBaseOutputGenerator.__init__(self)

    #
    # Called at beginning of processing as file is opened
    def generate(self):
        file_start = self.GenerateFileStart(os.path.basename(__file__))
        self.write(file_start)

        if self.filename == 'command_recorder.h':
            self.generateHeader()
        elif self.filename == 'command_recorder.cpp':
            self.generateSource()
        else:
            self.write(f'\nFile name {self.filename} has no code to generate\n')

        self.write(self.GenerateFileEnd())

    def generateHeader(self):
        out = []
        out.append('''
#include <cstring>
#include <iostream>
#include <vector>
#include <vulkan/vulkan.h>

#include "linear_allocator.h"
#include "command_common.h"

class CommandRecorder
{
  public:
  void Reset() { m_allocator.Reset(); }
''')
        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            func_decl = vkcommand.cPrototype.replace('VKAPI_ATTR ', '').replace('VKAPI_CALL ', '').replace('vk', 'Record', 1).replace(f'{vkcommand.returnType} ', f'{vkcommand.name[2:]}Args*', 1)
            out.append(f'  {func_decl}\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')
        out.append('''

  private:

    template <typename T> T *Alloc() { return new(m_allocator.Alloc(sizeof(T))) T; }
    template <typename T> T *CopyArray(const T *src, size_t start_index, size_t count) {
      auto ptr = reinterpret_cast<T *>(m_allocator.Alloc(sizeof(T) * count));
      std::memcpy(ptr, src, sizeof(T) * count);
      return ptr;
    }

    LinearAllocator<> m_allocator;
};
''')
        self.write("".join(out))

    def generateSource(self):
        out = []
        out.append('''
#include <iomanip>
#include <sstream>

#include "command_common.h"
#include "command_recorder.h"

// Declare CopyArray template functions. We need this declaration since the
// templates call each other and we don't have control over the order of the
// definitions.

''')
        # Find all the types that are used by command buffer related calls. These are the only
        # ones we need to record and skipping the rest of the spec avoids many tricky corner cases
        # and 20k lines of unused code.
        #
        cmd_types = set()

        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            for vkparam in vkcommand.params:
                cmd_types.add(vkparam.type)

        for vkstruct in self.vk.structs.values():
            if vkstruct.name not in cmd_types:
                continue
            for vkmember in vkstruct.members:
                cmd_types.add(vkmember.type)

        out.append('template<>\n')
        out.append('uint8_t* CommandRecorder::CopyArray<uint8_t>(const uint8_t* src, size_t start_index, size_t count);\n')
        for vkstruct in self.vk.structs.values():
            if vkstruct.name not in cmd_types:
                continue
            out.extend([f'#ifdef {vkstruct.protect}\n'] if vkstruct.protect else [])
            out.append('template<>\n')
            out.append(f'{vkstruct.name}* CommandRecorder::CopyArray<{vkstruct.name}>(const {vkstruct.name}* src, size_t start_index, size_t count);\n')
            out.extend([f'#endif //{vkstruct.protect}\n'] if vkstruct.protect else [])
        out.append('\n\n// Define CopyArray template functions.\n\n')

        out.append('template<>\n')
        out.append('uint8_t* CommandRecorder::CopyArray<uint8_t>(const uint8_t* src, size_t start_index, size_t count) {\n')
        out.append('    auto ptr = reinterpret_cast<uint8_t*>(m_allocator.Alloc(sizeof(uint8_t) * count));\n')
        out.append('    memcpy(ptr, src, sizeof(uint8_t) * count);\n')
        out.append('    return ptr;\n')
        out.append('}\n')
        for vkstruct in self.vk.structs.values():
            if vkstruct.name not in cmd_types:
                continue
            out.extend([f'#ifdef {vkstruct.protect}\n'] if vkstruct.protect else [])
            out.append('template<>\n')
            out.append(f'{vkstruct.name}* CommandRecorder::CopyArray<{vkstruct.name}>(const {vkstruct.name}* src, size_t start_index, size_t count) {{\n')
            out.append(f'  auto ptr = reinterpret_cast<{vkstruct.name}*>(m_allocator.Alloc(sizeof({vkstruct.name}) * count));\n')
            out.append('  for (uint64_t i = 0; i < count; ++i) {\n')
            src_struct = 'src[start_index + i]'
            for vkmember in vkstruct.members:
                if vkmember.fixedSizeArray is not None and len(vkmember.fixedSizeArray) > 0 and 'PFN_' not in vkmember.type:
                    out.append(f'  for (uint32_t j = 0; j < {vkmember.fixedSizeArray[0]}; ++j) {{\n')
                    out.append(f'    ptr[i].{vkmember.name}[j] = {src_struct}.{vkmember.name}[j];\n')
                    out.append('  }\n')
                # https://github.com/LunarG/CrashDiagnosticLayer/issues/102 we need to deep copy the pNext chain here
                elif vkmember.name == 'pNext':
                    out.append(f'  ptr[i].{vkmember.name} = nullptr; // pNext deep copy not implemented\n')
                elif vkmember.pointer and 'void' != vkmember.type:
                    out.append(f'    ptr[i].{vkmember.name} = nullptr;\n')
                    out.append(f'  if ({src_struct}.{vkmember.name}) {{\n')
                    if vkmember.length is not None and len(vkmember.length) > 0:
                        lengths = vkmember.length.split(',')
                        out.append(f'    ptr[i].{vkmember.name} = CopyArray({src_struct}.{vkmember.name}, 0U, {src_struct}.{lengths[0]});\n')
                        # TODO: this only handles 2d arrays with a constant 2nd dimension
                        # because that's all that is in vk.xml right now
                        if len(lengths) > 1:
                            out.append(f'for (uint64_t j = 0; j < uint64_t({src_struct}.{lengths[0]}); j++) {{\n')
                            out.append(f'    const_cast<{vkmember.type} **>(ptr[i].{vkmember.name})[j] = CopyArray({src_struct}.{vkmember.name}[j], 0U, {lengths[1]});\n')
                            out.append('}')

                    elif vkmember.type == 'char':
                        out.append(f'    ptr[i].{vkmember.name} = CopyArray<>({src_struct}.{vkmember.name}, 0, strlen({src_struct}.{vkmember.name}) + 1);\n')
                    else:
                        out.append(f'    ptr[i].{vkmember.name} = CopyArray({src_struct}.{vkmember.name}, 0U, 1U);\n')
                    out.append('  }\n')
                elif vkmember.length is not None and len(vkmember.length) > 0:
                    out_type = vkmember.cDeclaration.replace(vkmember.name, '').strip()
                    out.append(f'    ptr[i].{vkmember.name} = reinterpret_cast<{out_type}>(CopyArray(reinterpret_cast<const uint8_t*>({src_struct}.{vkmember.name}), 0U, {src_struct}.{vkmember.length}));\n')
                else:
                    out.append(f'  ptr[i].{vkmember.name} = {src_struct}.{vkmember.name};\n')
            out.append('  }\n')
            out.append('  return ptr;\n')
            out.append('}\n')
            out.extend([f'#endif //{vkstruct.protect}\n'] if vkstruct.protect else [])
            out.append('\n')

        for vkcommand in filter(lambda x: self.CommandBufferCall(x), self.vk.commands.values()):
            out.extend([f'#ifdef {vkcommand.protect}\n'] if vkcommand.protect else [])
            func_decl = vkcommand.cPrototype.replace('VKAPI_ATTR ', '').replace('VKAPI_CALL ', '').replace('vk', 'CommandRecorder::Record').replace(f'{vkcommand.returnType} ', f'{vkcommand.name[2:]}Args*', 1).replace(';', ' {', 1)
            out.append(f'{func_decl}\n')
            out.append(f'  auto *args = Alloc<{vkcommand.name[2:]}Args>();\n')
            for vkparam in vkcommand.params:
                if vkparam.fixedSizeArray is not None and len(vkparam.fixedSizeArray) > 0:
                    out.append(f'  for (uint32_t i = 0; i < {vkparam.length}; ++i) {{\n')
                    out.append(f'    args->{vkparam.name}[i] = {vkparam.name}[i];\n')
                    out.append('  }\n')
                elif vkparam.pointer and 'void' != vkparam.type:
                    out.append(f'  if ({vkparam.name}) {{\n')
                    if vkparam.length is not None and len(vkparam.length) > 0:
                        out.append(f'    args->{vkparam.name} = CopyArray({vkparam.name}, static_cast<size_t>(0U), static_cast<size_t>({vkparam.length}));\n')
                    else:
                        out.append(f'    args->{vkparam.name} = CopyArray({vkparam.name}, static_cast<size_t>(0U), static_cast<size_t>(1U));\n')
                    out.append('  }\n')
                elif vkparam.length is not None and len(vkparam.length) > 0:
                    out.append(f'    args->{vkparam.name} = CopyArray(reinterpret_cast<const uint8_t*>({vkparam.name}), static_cast<size_t>(0U), static_cast<size_t>({vkparam.length}));\n')
                else:
                    out.append(f'  args->{vkparam.name} = {vkparam.name};\n')
            out.append('  return args;\n')
            out.append('}\n')
            out.extend([f'#endif //{vkcommand.protect}\n'] if vkcommand.protect else [])
            out.append('\n')
        self.write("".join(out))



#!/usr/bin/env python3
# Copyright (c) 2021-2025 The Khronos Group Inc.
# Copyright (c) 2021-2025 Valve Corporation
# Copyright (c) 2021-2025 LunarG, Inc.
# Copyright (c) 2021-2024 Google Inc.
# Copyright (c) 2023-2024 RasterGrid Kft.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import filecmp
import os
import shutil
import subprocess
import sys
import tempfile
import difflib
import json
import common_ci
import pickle
from xml.etree import ElementTree

def RunGenerators(api: str, registry: str, grammar: str, directory: str, styleFile: str, targetFilter: str, caching: bool):

    has_clang_format = shutil.which('clang-format') is not None
    if not has_clang_format:
        print("WARNING: Unable to find clang-format!")

    # These live in the Vulkan-Docs repo, but are pulled in via the
    # Vulkan-Headers/registry folder
    # At runtime we inject python path to find these helper scripts
    scripts = os.path.dirname(registry)
    scripts_directory_path = os.path.dirname(os.path.abspath(__file__))
    registry_headers_path = os.path.join(scripts_directory_path, scripts)
    sys.path.insert(0, registry_headers_path)
    try:
        from reg import Registry
    except:
        print("ModuleNotFoundError: No module named 'reg'") # normal python error message
        print(f'{registry_headers_path} is not pointing to the Vulkan-Headers registry directory.')
        print("Inside Vulkan-Headers there is a registry/reg.py file that is used.")
        sys.exit(1) # Return without call stack so easy to spot error

    from base_generator import BaseGeneratorOptions
    from generators.command_common_generator import CommandCommonOutputGenerator
    from generators.command_printer_generator import CommandPrinterOutputGenerator
    from generators.command_recorder_generator import CommandRecorderOutputGenerator
    from generators.dispatch_generator import DispatchOutputGenerator
    from generators.command_prepost_generator import CommandPrePostGenerator
    from generators.command_tracker_generator import CommandTrackerGenerator
    from generators.context_generator import InterceptContextGenerator
    from generators.layer_base_generator import LayerBaseOutputGenerator
    from generators.test_icd_generator import TestIcdGenerator
    from generators.object_name_generator import ObjectNameGenerator

    # These set fields that are needed by both OutputGenerator and BaseGenerator,
    # but are uniform and don't need to be set at a per-generated file level
    from base_generator import SetOutputDirectory, SetTargetApiName, SetMergedApiNames, EnableCaching
    SetOutputDirectory(directory)
    SetTargetApiName(api)

    # Build up a list of all generators
    # Note: Options variable names MUST match order of constructor variable in generator
    generators = {
        'command_common.h' : {
            'generator': CommandCommonOutputGenerator,
            'genCombined': True,
        },
        'command_common.cpp' : {
            'generator': CommandCommonOutputGenerator,
            'genCombined': True,
        },
        'command_printer.h' : {
            'generator': CommandPrinterOutputGenerator,
            'genCombined': True,
        },
        'command_printer.cpp' : {
            'generator': CommandPrinterOutputGenerator,
            'genCombined': True,
        },
        'command_printer_structs.cpp' : {
            'generator': CommandPrinterOutputGenerator,
            'genCombined': True,
        },
        'command_printer_types.cpp' : {
            'generator': CommandPrinterOutputGenerator,
            'genCombined': True,
        },
        'command_recorder.h' : {
            'generator': CommandRecorderOutputGenerator,
            'genCombined': True,
        },
        'command_recorder.cpp' : {
            'generator': CommandRecorderOutputGenerator,
            'genCombined': True,
        },
        'command.h.inc' : {
            'generator': CommandPrePostGenerator,
            'genCombined': True,
        },
        'command.cpp.inc' : {
            'generator': CommandPrePostGenerator,
            'genCombined': True,
        },
        'command_tracker.h' : {
            'generator': CommandTrackerGenerator,
            'genCombined': True,
        },
        'command_tracker.cpp' : {
            'generator': CommandTrackerGenerator,
            'genCombined': True,
        },
        'dispatch.h' : {
            'generator': DispatchOutputGenerator,
            'genCombined': True,
        },
        'dispatch.cpp' : {
            'generator': DispatchOutputGenerator,
            'genCombined': True,
        },
        'cdl_commands.h.inc' : {
            'generator': InterceptContextGenerator,
            'genCombined': True,
        },
        'cdl_commands.cpp.inc' : {
            'generator': InterceptContextGenerator,
            'genCombined': True,
        },
        'layer_base.h.inc' : {
            'generator': LayerBaseOutputGenerator,
            'genCombined': True,
        },
        'layer_base.cpp.inc' : {
            'generator': LayerBaseOutputGenerator,
            'genCombined': True,
        },
        'test_icd_helper.h' : {
            'generator': TestIcdGenerator,
            'genCombined': True,
        },
        'object_name.h' : {
            'generator': ObjectNameGenerator,
            'genCombined': True,
        },
    }

    unknownTargets = [x for x in (targetFilter if targetFilter else []) if x not in generators.keys()]
    if unknownTargets:
        print(f'ERROR: No generator options for unknown target(s): {", ".join(unknownTargets)}', file=sys.stderr)
        return 1

    # Filter if --target is passed in
    targets = [x for x in generators.keys() if not targetFilter or x in targetFilter]

    cacheVkObjectData = None
    cachePath = os.path.join(tempfile.gettempdir(), f'vkobject_{os.getpid()}')
    if caching:
        EnableCaching()

    for index, target in enumerate(targets, start=1):
        print(f'[{index}|{len(targets)}] Generating {target}')

        # First grab a class contructor object and create an instance
        targetGenerator = generators[target]['generator']
        generatorOptions = generators[target]['options'] if 'options' in generators[target] else []
        generator = targetGenerator(*generatorOptions)

        # This code and the 'genCombined' generator metadata is used by downstream
        # users to generate code with all Vulkan APIs merged into the target API variant
        # (e.g. Vulkan SC) when needed. The constructed apiList is also used to filter
        # out non-applicable extensions later below.
        apiList = [api]
        if api != 'vulkan' and generators[target]['genCombined']:
            SetMergedApiNames('vulkan')
            apiList.append('vulkan')
        else:
            SetMergedApiNames(None)

        baseOptions = BaseGeneratorOptions(customFileName = target)

        # Create the registry object with the specified generator and generator
        # options. The options are set before XML loading as they may affect it.
        reg = Registry(generator, baseOptions)

        # Parse the specified registry XML into an ElementTree object
        tree = ElementTree.parse(registry)

        # Filter out extensions that are not on the API list
        [exts.remove(e) for exts in tree.findall('extensions') for e in exts.findall('extension') if (sup := e.get('supported')) is not None and all(api not in sup.split(',') for api in apiList)]

        # Load the XML tree into the registry object
        reg.loadElementTree(tree)

        # The cached data is saved inside the BaseGenerator, so search for it and try
        # to reuse the parsing for each generator file.
        if caching and not cacheVkObjectData:
            if os.path.isfile(cachePath):
                file = open(cachePath, 'rb')
                cacheVkObjectData = pickle.load(file)
                file.close()

        if caching and cacheVkObjectData and 'regenerate' not in generators[target]:
            # TODO - We shouldn't have to regenerate any files, need to investigate why we some scripts need it
            reg.gen.generateFromCache(cacheVkObjectData, reg.genOpts)
        else:
            # Finally, use the output generator to create the requested target
            reg.apiGen()

        # Run clang-format on the file
        if has_clang_format:
            common_ci.RunShellCmd(f'clang-format -i --style=file:{styleFile} {os.path.join(directory, target)}')

    if os.path.isfile(cachePath):
        os.remove(cachePath)

# helper to define paths relative to the repo root
def repo_relative(path):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), '..', path))

def main(argv):
    # files to exclude from --verify check
    verify_exclude = [
        '.clang-format',
    ]

    parser = argparse.ArgumentParser(description='Generate source code for this repository')
    parser.add_argument('--api',
                        default='vulkan',
                        choices=['vulkan'],
                        help='Specify API name to generate')
    parser.add_argument('paths', nargs='+',
                        help='Either: Paths to the Vulkan-Headers registry directory and the SPIRV-Headers grammar directory'
                        + ' OR path to the base directory containing the Vulkan-Headers and SPIRV-Headers repositories')
    parser.add_argument('--generated-version', help='sets the header version used to generate the repo')
    parser.add_argument('-o', help='Create target and related files in specified directory.', dest='output_directory')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--target', nargs='+', help='only generate file names passed in')
    group.add_argument('-i', '--incremental', action='store_true', help='only update repo files that change')
    group.add_argument('-v', '--verify', action='store_true', help='verify repo files match generator output')
    group.add_argument('--no-caching', action='store_true', help='Do not try to cache generator objects')
    args = parser.parse_args(argv)

    repo_dir = repo_relative(f'src/generated')

    # Need pass style file incase running with --verify and it can't find the file automatically in the temp directory
    styleFile = repo_relative('.clang-format')
    if common_ci.IsGHA() and args.verify:
        # Have found that sometimes (~5%) the 20.04 Ubuntu machines have clang-format v11 but we need v14 to
        # use a dedicated styleFile location. For these case there we can survive just skipping the verify check
        stdout = subprocess.check_output(['clang-format', '--version']).decode("utf-8")
        version = stdout[stdout.index('version') + 8:][:2]
        if int(version) < 14:
            return 0 # Success

    # Update the api_version in the respective json files
    if args.generated_version:
        json_files = []
        json_files.append(repo_relative('src/crash_diagnostic_layer.json.in'))
        for json_file in json_files:
            with open(json_file) as f:
                data = json.load(f)

            data["layer"]["api_version"] = args.generated_version

            with open(json_file, mode='w', encoding='utf-8', newline='\n') as f:
                f.write(json.dumps(data, indent=4))

    # get directory where generators will run
    if args.verify or args.incremental:
        # generate in temp directory so we can compare or copy later
        temp_obj = tempfile.TemporaryDirectory(prefix='cdl_codegen_')
        temp_dir = temp_obj.name
        gen_dir = temp_dir
    else:
        # generate directly in the repo
        gen_dir = repo_dir

    if args.output_directory is not None:
      gen_dir = args.output_directory

    if len(args.paths) == 1:
        base = args.paths[0]
        registry = os.path.join(base, 'Vulkan-Headers/registry')
        grammar = os.path.join(base, 'SPIRV-Headers/include/spirv/unified1')
    elif len(args.paths) == 2:
        registry = args.paths[0]
        grammar = args.paths[1]
    else:
        args.print_help()
        return -1

    registry = os.path.abspath(os.path.join(registry,  'vk.xml'))
    if not os.path.isfile(registry):
        print(f'{registry} does not exist')
        return -1
    grammar = os.path.abspath(os.path.join(grammar, 'spirv.core.grammar.json'))
    if not os.path.isfile(grammar):
        print(f'{grammar} does not exist')
        return -1

    caching = not args.no_caching
    RunGenerators(args.api, registry, grammar, gen_dir, styleFile, args.target, caching)

    # optional post-generation steps
    if args.verify:
        # compare contents of temp dir and repo
        temp_files = set(os.listdir(temp_dir))
        repo_files = set(os.listdir(repo_dir))
        for filename in sorted((temp_files | repo_files) - set(verify_exclude)):
            temp_filename = os.path.join(temp_dir, filename)
            repo_filename = os.path.join(repo_dir, filename)
            if filename not in repo_files:
                print('ERROR: Missing repo file', filename)
                return 2
            elif filename not in temp_files:
                print('ERROR: Missing generator for', filename)
                return 3
            elif not filecmp.cmp(temp_filename, repo_filename, shallow=False):
                print('ERROR: Repo files do not match generator output for', filename)
                # print line diff on file mismatch
                with open(temp_filename) as temp_file, open(repo_filename) as repo_file:
                    print(''.join(difflib.unified_diff(temp_file.readlines(),
                                                       repo_file.readlines(),
                                                       fromfile='temp/' + filename,
                                                       tofile=  'repo/' + filename)))
                return 4

        # return code for test scripts
        print('SUCCESS: Repo files match generator output')

    elif args.incremental:
        # copy missing or differing files from temp directory to repo
        for filename in os.listdir(temp_dir):
            temp_filename = os.path.join(temp_dir, filename)
            repo_filename = os.path.join(repo_dir, filename)
            if not os.path.exists(repo_filename) or \
               not filecmp.cmp(temp_filename, repo_filename, shallow=False):
                print('update', repo_filename)
                shutil.copyfile(temp_filename, repo_filename)

    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))




#!/usr/bin/python3 -i
#
# Copyright (c) 2023-2024 Valve Corporation
# Copyright (c) 2023-2024 LunarG, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import json

# Build a set of all vuid text strings found in validusage.json
def buildListVUID(valid_usage_file: str) -> set:

    # Walk the JSON-derived dict and find all "vuid" key values
    def ExtractVUIDs(vuid_dict):
        if hasattr(vuid_dict, 'items'):
            for key, value in vuid_dict.items():
                if key == "vuid":
                    yield value
                elif isinstance(value, dict):
                    for vuid in ExtractVUIDs(value):
                        yield vuid
                elif isinstance (value, list):
                    for listValue in value:
                        for vuid in ExtractVUIDs(listValue):
                            yield vuid

    valid_vuids = set()
    if not os.path.isfile(valid_usage_file):
        print(f'Error: Could not find, or error loading {valid_usage_file}')
        sys.exit(1)
    json_file = open(valid_usage_file, 'r', encoding='utf-8')
    vuid_dict = json.load(json_file)
    json_file.close()
    if len(vuid_dict) == 0:
        print(f'Error: Failed to load {valid_usage_file}')
        sys.exit(1)
    for json_vuid_string in ExtractVUIDs(vuid_dict):
        valid_vuids.add(json_vuid_string)

    # List of VUs that should exists, but have a spec bug
    for vuid in [
        # https://gitlab.khronos.org/vulkan/vulkan/-/issues/3582
        "VUID-VkCopyImageToImageInfoEXT-commonparent",
        "VUID-vkUpdateDescriptorSetWithTemplate-descriptorSet-parent",
        "VUID-vkUpdateVideoSessionParametersKHR-videoSessionParameters-parent",
        "VUID-vkDestroyVideoSessionParametersKHR-videoSessionParameters-parent",
        "VUID-vkGetDescriptorSetHostMappingVALVE-descriptorSet-parent",
        ]:
        valid_vuids.add(vuid)

    return valid_vuids

# Will do a sanity check the VUID exists
def getVUID(valid_vuids: set, vuid: str, quotes: bool = True) -> str:
    if vuid not in valid_vuids:
        print(f'Warning: Could not find {vuid} in validusage.json')
        vuid = vuid.replace('VUID-', 'UNASSIGNED-')
    return vuid if not quotes else f'"{vuid}"'

class PlatformGuardHelper():
    """Used to elide platform guards together, so redundant #endif then #ifdefs are removed
    Note - be sure to call add_guard(None) when done to add a trailing #endif if needed
    """
    def __init__(self):
        self.current_guard = None

    def add_guard(self, guard, extra_newline = False):
        out = []
        if self.current_guard != guard and self.current_guard is not None:
            out.append(f'#endif  // {self.current_guard}\n')
        if extra_newline:
            out.append('\n')
        if self.current_guard != guard and guard is not None:
            out.append(f'#ifdef {guard}\n')
        self.current_guard = guard
        return out

# The SPIR-V grammar json doesn't have an easy way to detect these, so have listed by hand
# If we are missing one, its not critical, the goal of this list is to reduce the generated output size
def IsNonVulkanSprivCapability(capability):
    return capability in [
        'Kernel',
        'Vector16',
        'Float16Buffer',
        'ImageBasic',
        'ImageReadWrite',
        'ImageMipmap',
        'DeviceEnqueue',
        'SubgroupDispatch',
        'Pipes',
        'LiteralSampler',
        'NamedBarrier',
        'PipeStorage',
        'SubgroupShuffleINTEL',
        'SubgroupShuffleINTEL',
        'SubgroupBufferBlockIOINTEL',
        'SubgroupImageBlockIOINTEL',
        'SubgroupImageMediaBlockIOINTEL',
        'RoundToInfinityINTEL',
        'FloatingPointModeINTEL',
        'IndirectReferencesINTEL',
        'AsmINTEL',
        'VectorComputeINTEL',
        'VectorAnyINTEL',
        'SubgroupAvcMotionEstimationINTEL',
        'SubgroupAvcMotionEstimationIntraINTEL',
        'SubgroupAvcMotionEstimationChromaINTEL',
        'VariableLengthArrayINTEL',
        'FunctionFloatControlINTEL',
        'FPGAMemoryAttributesINTEL',
        'FPFastMathModeINTEL',
        'ArbitraryPrecisionIntegersINTEL',
        'ArbitraryPrecisionFloatingPointINTEL',
        'UnstructuredLoopControlsINTEL',
        'FPGALoopControlsINTEL',
        'KernelAttributesINTEL',
        'FPGAKernelAttributesINTEL',
        'FPGAMemoryAccessesINTEL',
        'FPGAClusterAttributesINTEL',
        'LoopFuseINTEL',
        'FPGADSPControlINTEL',
        'MemoryAccessAliasingINTEL',
        'FPGAInvocationPipeliningAttributesINTEL',
        'FPGABufferLocationINTEL',
        'ArbitraryPrecisionFixedPointINTEL',
        'USMStorageClassesINTEL',
        'RuntimeAlignedAttributeINTEL',
        'IOPipesINTEL',
        'BlockingPipesINTEL',
        'FPGARegINTEL',
        'LongCompositesINTEL',
        'OptNoneINTEL',
        'DebugInfoModuleINTEL',
        'BFloat16ConversionINTEL',
        'SplitBarrierINTEL',
        'FPGAClusterAttributesV2INTEL',
        'FPGAKernelAttributesv2INTEL',
        'FPMaxErrorINTEL',
        'FPGALatencyControlINTEL',
        'FPGAArgumentInterfacesINTEL',
        'GlobalVariableHostAccessINTEL',
        'GlobalVariableFPGADecorationsINTEL',
        'MaskedGatherScatterINTEL',
        'CacheControlsINTEL',
        'RegisterLimitsINTEL'
    ]



#!/usr/bin/python3 -i
#
# Copyright 2023-2025 The Khronos Group Inc.
# Copyright 2023-2025 Valve Corporation
# Copyright 2023-2025 LunarG, Inc.
#
# SPDX-License-Identifier: Apache-2.0

import os
from base_generator import BaseGenerator
from generators.generator_utils import PlatformGuardHelper

class ObjectNameGenerator(BaseGenerator):
    def __init__(self):
        BaseGenerator.__init__(self)

    def generate(self):
        out = []
        out.append(f'''// *** THIS FILE IS GENERATED - DO NOT EDIT ***
// See {os.path.basename(__file__)} for modifications
// Copyright 2023-2025 The Khronos Group Inc.
// Copyright 2023-2025 Valve Corporation
// Copyright 2023-2025 LunarG, Inc.
//
// SPDX-License-Identifier: Apache-2.0
''')
        out.append('''
#pragma once

#ifdef __cplusplus
#include <string>
#endif
#include <vulkan/vulkan.h>
''')
        guard_helper = PlatformGuardHelper()
        out.append(f'static inline const char* string_ObjectName(VkObjectType input_value) {{\n')
        out.append('    switch (input_value) {\n')
        for handle in self.vk.handles.values():
            out.extend(guard_helper.add_guard(handle.protect))
            out.append(f'        case {handle.type}:\n')
            out.append(f'            return "{handle.name}";\n')
        out.extend(guard_helper.add_guard(None))
        out.append('        default:\n')
        out.append(f'            return "UNKNOWN";\n')
        out.append('    }\n')
        out.append('}\n')
        out.append('\n')

        self.write("".join(out))
